# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# ----------------------------------------------------------------------------

"""
Tests for explorer plugin utilities.
"""

# Standard imports
import os
import sys

# Third party imports
from qtpy.QtCore import Qt, QTimer
import pytest

# Local imports
from spyder.plugins.explorer.widgets.fileassociations import (
    ApplicationsDialog, InputTextDialog)


@pytest.mark.order(1)
def test_input_text_dialog(qtbot):
    widget = InputTextDialog()
    qtbot.addWidget(widget)
    widget.show()

    # Test empty
    widget.validate()
    assert not widget.button_ok.isEnabled()

    # Test text
    widget.set_text('hello')
    widget.validate()
    assert widget.button_ok.isEnabled()

    # Test regex ok enabled
    widget.set_text('')
    widget.set_regex_validation('hello')
    qtbot.keyClicks(widget.lineedit, 'hello world!')
    assert widget.text() == 'hello'
    assert widget.button_ok.isEnabled()
    widget.validate()

    # Test regex ok disabled
    widget.set_text('')
    widget.set_regex_validation('hello')
    qtbot.keyClicks(widget.lineedit, 'hell')
    assert not widget.button_ok.isEnabled()
    widget.validate()


@pytest.mark.order(1)
def test_apps_dialog(qtbot, tmp_path):
    widget = ApplicationsDialog()
    qtbot.addWidget(widget)
    widget.show()

    if os.name == 'nt':
        ext = '.exe'
    elif sys.platform == 'darwin':
        ext = '.app'
    else:
        ext = '.desktop'

    mock_apps = {
        'some app 1': '/some/fake/some app 1' + ext,
        'some app 2': '/some/fake/path/some app 2' + ext,
        'some app 3': '/some/fake/path/some app 3' + ext,
    }
    widget.setup(mock_apps)

    # Test filter
    qtbot.keyClicks(widget.edit_filter, '1')
    count_hidden = 0
    for row in range(widget.list.count()):
        item = widget.list.item(row)
        count_hidden += int(item.isHidden())
    assert count_hidden == 2

    # Test selected app
    widget.list.setCurrentItem(widget.list.item(0))
    assert widget.application_name == 'some app 1'
    assert widget.application_path == '/some/fake/some app 1' + ext

    # Test extension label
    widget.set_extension('.hello')
    assert '.hello' in widget.label.text()

    # Test reset filter
    widget.edit_filter.setText('')
    count_hidden = 0
    for row in range(widget.list.count()):
        item = widget.list.item(row)
        count_hidden += int(item.isHidden())
    assert count_hidden == 0

    # Test browse invalid
    fpath = '/some/other/path'
    widget.browse(fpath)
    assert widget.list.count() == 3

    # Test browse valid
    if os.name == 'nt':
        path_obj = tmp_path / ("some-new-app" + ext)
        path_obj.write_bytes(b'\x00\x00')
        fpath = str(path_obj)
    elif sys.platform == 'darwin':
        path_obj = tmp_path / ("some-new-app" + ext)
        path_obj.mkdir()
        fpath = str(path_obj)
    else:
        path_obj = tmp_path / ("some-new-app" + ext)
        path_obj.write_text(u'''
[Desktop Entry]
Name=Suer app
Type=Application
Exec=/something/bleerp
Icon=/blah/blah.xpm
''')
        fpath = str(path_obj)

    widget.browse(fpath)
    assert widget.list.count() == 4

    # Test browse valid duplicate
    widget.browse(fpath)
    assert widget.list.count() == 4


def create_timer(func, interval=500):
    """Helper function to help interact with modal dialogs."""
    timer = QTimer()
    timer.setInterval(interval)
    timer.setSingleShot(True)
    timer.timeout.connect(func)
    timer.start()
    return timer


@pytest.mark.order(1)
def test_file_assoc_widget(file_assoc_widget):
    qtbot, widget = file_assoc_widget

    # Test data
    assert widget.data == widget.test_data

    # Test add invalid associations
    extension = 'blooper.foo,'

    def interact_with_dialog_1():
        qtbot.keyClicks(widget._dlg_input.lineedit, extension)
        assert widget._dlg_input.lineedit.text() == extension
        assert not widget._dlg_input.button_ok.isEnabled()
        qtbot.keyClick(widget._dlg_input.button_cancel, Qt.Key_Return)

    _ = create_timer(interact_with_dialog_1)
    qtbot.mouseClick(widget.button_add, Qt.LeftButton)

    # Test add valid association
    extension = '*.zpam,MANIFEST.in'

    def interact_with_dialog_2():
        qtbot.keyClicks(widget._dlg_input.lineedit, extension)
        qtbot.keyClick(widget._dlg_input.button_ok, Qt.Key_Return)

    _ = create_timer(interact_with_dialog_2)
    qtbot.mouseClick(widget.button_add, Qt.LeftButton)
    assert widget.list_extensions.count() == 3
    assert widget.list_extensions.item(2).text() == extension

    # Test add invalid association programmatically
    widget.add_association(value='mehh')
    assert widget.list_extensions.count() == 3

    # Test add valid association programmatically
    widget.add_association(value='*.boom')
    assert widget.list_extensions.count() == 4

    # Test add repeated association programmatically
    widget.add_association(value='*.csv')
    assert widget.list_extensions.count() == 4
    widget._add_association(value='*.csv')
    assert widget.list_extensions.count() == 4

    # Test edit association
    extension = '*.zpam'

    def interact_with_dialog_3():
        widget._dlg_input.lineedit.clear()
        qtbot.keyClicks(widget._dlg_input.lineedit, extension)
        qtbot.keyClick(widget._dlg_input.button_ok, Qt.Key_Return)

    _ = create_timer(interact_with_dialog_3)
    qtbot.mouseClick(widget.button_edit, Qt.LeftButton)
    assert widget.list_extensions.count() == 4
    assert widget.list_extensions.item(2).text() == extension

    # Test remove associations
    qtbot.mouseClick(widget.button_remove, Qt.LeftButton)
    assert widget.list_extensions.count() == 3

    # Test set default
    widget.list_applications.setCurrentRow(1)
    qtbot.mouseClick(widget.button_default, Qt.LeftButton)
    assert 'App name 2' in widget.list_applications.item(0).text()

    # Test add application
    def interact_with_dialog_4():
        assert not widget._dlg_applications.button_ok.isEnabled()
        count = widget._dlg_applications.list.count()
        if count > 0:
            widget._dlg_applications.list.setCurrentRow(count - 1)
            qtbot.keyClick(widget._dlg_applications.button_ok, Qt.Key_Return)
        else:
            qtbot.keyClick(widget._dlg_applications.button_cancel,
                           Qt.Key_Return)

    _ = create_timer(interact_with_dialog_4)
    qtbot.mouseClick(widget.button_add_application, Qt.LeftButton)
    count = widget.list_applications.count()
    assert count in [2, 3]

    # Test add repeated application programmatically
    app_name, app_path = widget.test_data['*.csv'][0]
    widget._add_application(app_name, app_path)
    count = widget.list_applications.count()
    assert count in [2, 3]

    # Test remove application
    widget.list_applications.setCurrentRow(0)
    qtbot.mouseClick(widget.button_remove_application, Qt.LeftButton)
    count = widget.list_applications.count()
    assert count in [1, 2]
    assert 'App name 1' in widget.list_applications.item(0).text()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Module checking Spyder runtime dependencies"""

# Standard library imports
import sys

# Third party imports
from qtpy.QtGui import QColor
from qtpy.QtWidgets import (QApplication, QDialog, QDialogButtonBox,
                            QHBoxLayout, QVBoxLayout, QLabel, QPushButton,
                            QStackedWidget, QTreeWidget, QTreeWidgetItem)

# Local imports
from spyder import __version__
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.config.base import _
from spyder.config.gui import is_dark_interface
from spyder.dependencies import OPTIONAL, PLUGIN
from spyder.utils.icon_manager import ima
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle, MAC, WIN
from spyder.widgets.helperwidgets import PaneEmptyWidget


class DependenciesTreeWidget(QTreeWidget):

    def update_dependencies(self, dependencies):
        self.clear()
        headers = (_("Module"), _("Package name"), _(" Required "),
                   _(" Installed "), _("Provided features"))
        self.setHeaderLabels(headers)

        # Mandatory items
        mandatory_item = QTreeWidgetItem([_("Mandatory")])
        font = mandatory_item.font(0)
        font.setBold(True)
        mandatory_item.setFont(0, font)

        # Optional items
        optional_item = QTreeWidgetItem([_("Optional")])
        optional_item.setFont(0, font)

        # Spyder plugins
        spyder_plugins = QTreeWidgetItem([_("Spyder plugins")])
        spyder_plugins.setFont(0, font)

        self.addTopLevelItems([mandatory_item, optional_item, spyder_plugins])

        for dependency in sorted(dependencies,
                                 key=lambda x: x.modname.lower()):
            item = QTreeWidgetItem([dependency.modname,
                                    dependency.package_name,
                                    dependency.required_version,
                                    dependency.installed_version,
                                    dependency.features])

            # Format content
            if dependency.check():
                item.setIcon(0, ima.icon('dependency_ok'))
            elif dependency.kind == OPTIONAL:
                item.setIcon(0, ima.icon('dependency_warning'))
                item.setBackground(2, QColor(SpyderPalette.COLOR_WARN_1))

                # Fix foreground color in the light theme
                if not is_dark_interface():
                    item.setForeground(
                        2, QColor(SpyderPalette.COLOR_BACKGROUND_1)
                    )
            else:
                item.setIcon(0, ima.icon('dependency_error'))
                item.setBackground(2, QColor(SpyderPalette.COLOR_ERROR_1))

                # Fix foreground color in the light theme
                if not is_dark_interface():
                    item.setForeground(
                        2, QColor(SpyderPalette.COLOR_BACKGROUND_1)
                    )

            # Add to tree
            if dependency.kind == OPTIONAL:
                optional_item.addChild(item)
            elif dependency.kind == PLUGIN:
                spyder_plugins.addChild(item)
            else:
                mandatory_item.addChild(item)

        self.expandAll()

    def resize_columns_to_contents(self):
        for col in range(self.columnCount()):
            self.resizeColumnToContents(col)


class DependenciesDialog(QDialog):

    def __init__(self, parent):
        QDialog.__init__(self, parent)

        # Widgets
        note1 = _(
            "Optional modules are not required to run Spyder but enhance "
            "its functions."
        )

        note2 = _(
            "New dependencies or changed ones will be correctly detected only "
            "after Spyder is restarted."
        )

        notes_vmargin = "0.4em" if WIN else "0.3em"
        label = QLabel(
            (
                "<style>"
                "ul, li {{margin-left: -15px}}"
                "li {{margin-bottom: {}}}"
                "</style>"
                "<ul>"
                "<li>{}</li>"
                "<li>{}</li>"
                "</ul>"
            ).format(notes_vmargin, note1, note2)
        )

        self.treewidget = DependenciesTreeWidget(self)
        self.copy_btn = QPushButton(_("Copy to clipboard"))
        ok_btn = SpyderDialogButtonBox(QDialogButtonBox.Ok)

        # Widget setup
        self.setWindowTitle(
            _("Dependencies for Spyder {}").format(__version__)
        )
        self.setModal(False)
        self.copy_btn.setEnabled(False)

        # Create a QStackedWidget
        self.stacked_widget = QStackedWidget()

        # Create a loading message
        self.loading_pane = PaneEmptyWidget(
            self,
            "dependencies",
            _("Dependency information will be retrieved shortly. "
              "Please wait..."),
            bottom_stretch=1,
            spinner=True,
        )

        # Add the loading label and tree widget to the stacked widget
        self.stacked_widget.addWidget(self.loading_pane)
        self.stacked_widget.addWidget(self.treewidget)

        # Make sure the loading label is the one shown initially
        self.stacked_widget.setCurrentWidget(self.loading_pane)

        # Layout
        hlayout = QHBoxLayout()
        hlayout.addWidget(self.copy_btn)
        hlayout.addStretch()
        hlayout.addWidget(ok_btn)

        vlayout = QVBoxLayout()
        vlayout.setContentsMargins(*((5 * AppStyle.MarginSize,) * 4))
        vlayout.addWidget(self.stacked_widget)
        vlayout.addSpacing(AppStyle.MarginSize)
        vlayout.addWidget(label)
        vlayout.addSpacing((-2 if MAC else 1) * AppStyle.MarginSize)
        vlayout.addLayout(hlayout)

        self.setLayout(vlayout)
        self.setFixedSize(860, 560)

        # Signals
        self.copy_btn.clicked.connect(self.copy_to_clipboard)
        ok_btn.accepted.connect(self.accept)

    def set_data(self, dependencies):
        self.treewidget.update_dependencies(dependencies)
        self.treewidget.resize_columns_to_contents()

        # Once data is loaded, switch to the tree widget
        self.stacked_widget.setCurrentWidget(self.treewidget)

        # Enable copy button
        self.copy_btn.setEnabled(True)

    def copy_to_clipboard(self):
        from spyder.dependencies import status

        QApplication.clipboard().setText(status())


def test():
    """Run dependency widget test"""
    from spyder import dependencies

    # Test sample
    dependencies.add("IPython", "IPython", "Enhanced Python interpreter",
                     ">=20.0")
    dependencies.add("matplotlib", "matplotlib", "Interactive data plotting",
                     ">=1.0")
    dependencies.add("sympy", "sympy", "Symbolic Mathematics", ">=10.0",
                     kind=OPTIONAL)
    dependencies.add("foo", "foo", "Non-existent module", ">=1.0")
    dependencies.add("numpy", "numpy",  "Edit arrays in Variable Explorer",
                     ">=0.10", kind=OPTIONAL)

    from spyder.utils.qthelpers import qapplication
    app = qapplication()  # noqa
    dlg = DependenciesDialog(None)
    dlg.set_data(dependencies.DEPENDENCIES)
    dlg.show()
    sys.exit(dlg.exec_())


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-

# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

import os
from textwrap import dedent

import pytest
from qtpy.QtCore import QObject, Signal, Slot

from spyder.plugins.completion.api import (
    CompletionRequestTypes, WorkspaceUpdateKind)


class CompletionManager(QObject):
    """Dummy completion plugin that can handle LSP responses."""
    sig_response = Signal(str, dict)

    @Slot(str, dict)
    def handle_response(self, method, params):
        self.sig_response.emit(method, params)


@pytest.fixture(scope='module', params=[
    pytest.lazy_fixture('lsp_provider'),
    pytest.lazy_fixture('lsp_stdio_provider')])
def lsp_client_and_completion(request):
    """Create an LSP client/completion pair."""
    completion = CompletionManager()
    client = request.param.clients['python']['instance']
    return client, completion


@pytest.mark.order(3)
def test_didOpen(lsp_client_and_completion, qtbot):
    client, completion = lsp_client_and_completion

    # Parameters to perform a textDocument/didOpen request
    params = {
        'file': 'test.py',
        'language': 'python',
        'version': 1,
        'text': "",
        'response_callback': completion.handle_response,
        'codeeditor': completion,
        'requires_response': False
    }

    # Wait for the client to be started
    with qtbot.waitSignal(completion.sig_response, timeout=30000) as blocker:
        client.perform_request(CompletionRequestTypes.DOCUMENT_DID_OPEN,
                               params)
    response, _ = blocker.args

    # Assert the response has what we expect
    assert response == 'textDocument/publishDiagnostics'


@pytest.mark.order(3)
def test_get_signature(lsp_client_and_completion, qtbot):
    client, completion = lsp_client_and_completion

    # Parameters to perform a textDocument/didChange request
    params = {
        'file': 'test.py',
        'language': 'python',
        'version': 1,
        'text': "import os\nos.walk(\n",
        'codeeditor': completion,
        'requires_response': False
    }

    # Perform the request
    with qtbot.waitSignal(completion.sig_response, timeout=30000):
        client.perform_request(CompletionRequestTypes.DOCUMENT_DID_CHANGE,
                               params)

    # Parameters to perform a textDocument/signatureHelp request
    signature_params = {
        'file': 'test.py',
        'line': 1,
        'column': 10,
        'requires_response': True,
        'response_callback': completion.handle_response
    }

    # Perform the request
    with qtbot.waitSignal(completion.sig_response, timeout=30000) as blocker:
        client.perform_request(CompletionRequestTypes.DOCUMENT_SIGNATURE,
                               signature_params)
    _, response = blocker.args

    # Assert the response has what we expect
    assert response['params']['signatures']['label'].startswith('walk')


@pytest.mark.order(3)
def test_get_completions(lsp_client_and_completion, qtbot):
    client, completion = lsp_client_and_completion

    # Parameters to perform a textDocument/didChange request
    params = {
        'file': 'test.py',
        'language': 'python',
        'version': 1,
        'text': "import o",
        'codeeditor': completion,
        'requires_response': False
    }

    # Perform the request
    with qtbot.waitSignal(completion.sig_response, timeout=30000):
        client.perform_request(CompletionRequestTypes.DOCUMENT_DID_CHANGE,
                               params)

    # Parameters to perform a textDocument/completion request
    completion_params = {
        'file': 'test.py',
        'line': 0,
        'column': 8,
        'requires_response': True,
        'response_callback': completion.handle_response
    }

    # Perform the request
    with qtbot.waitSignal(completion.sig_response, timeout=30000) as blocker:
        client.perform_request(CompletionRequestTypes.DOCUMENT_COMPLETION,
                               completion_params)
    _, response = blocker.args

    # Assert the response has what we expect
    completions = response['params']
    assert 'os' in [x['label'] for x in completions]


@pytest.mark.order(3)
def test_go_to_definition(lsp_client_and_completion, qtbot):
    client, completion = lsp_client_and_completion

    # Parameters to perform a textDocument/didChange request
    params = {
        'file': 'test.py',
        'language': 'python',
        'version': 1,
        'text': "import os\nos.walk\n",
        'codeeditor': completion,
        'requires_response': False
    }

    # Perform the request
    with qtbot.waitSignal(completion.sig_response, timeout=30000):
        client.perform_request(CompletionRequestTypes.DOCUMENT_DID_CHANGE,
                               params)

    # Parameters to perform a textDocument/definition request
    go_to_definition_params = {
        'file': 'test.py',
        'line': 0,
        'column': 19,
        'requires_response': True,
        'response_callback': completion.handle_response
    }

    # Perform the request
    with qtbot.waitSignal(completion.sig_response, timeout=30000) as blocker:
        client.perform_request(CompletionRequestTypes.DOCUMENT_DEFINITION,
                               go_to_definition_params)
    _, response = blocker.args

    # Assert the response has what we expect
    definition = response['params']
    assert 'os.py' in definition['file']


@pytest.mark.order(3)
def test_local_signature(lsp_client_and_completion, qtbot):
    client, completion = lsp_client_and_completion

    # Parameters to perform a textDocument/didOpen request
    text = dedent('''
    def test(a, b):
        """Test docstring"""
        pass
    test''')
    params = {
        'file': 'test.py',
        'language': 'python',
        'version': 1,
        'text': text,
        'codeeditor': completion,
        'requires_response': False
    }

    # Perform the request
    with qtbot.waitSignal(completion.sig_response, timeout=30000) as blocker:
        client.perform_request(CompletionRequestTypes.DOCUMENT_DID_CHANGE,
                               params)

    # Parameters to perform a textDocument/hover request
    signature_params = {
        'file': 'test.py',
        'line': 4,
        'column': 0,
        'requires_response': True,
        'response_callback': completion.handle_response
    }

    # Perform the request
    with qtbot.waitSignal(completion.sig_response, timeout=30000) as blocker:
        client.perform_request(CompletionRequestTypes.DOCUMENT_HOVER,
                               signature_params)
    _, response = blocker.args

    # Assert the response has what we expect
    definition = response['params']
    assert 'Test docstring' in definition


@pytest.mark.order(3)
def test_send_workspace_folders_change(lsp_client_and_completion, qtbot):
    client, completion = lsp_client_and_completion
    folder = '/tmp/'

    # Test addition
    params = {
        'folder': folder,
        'kind': WorkspaceUpdateKind.ADDITION,
        'instance': ''
    }

    client.send_workspace_folders_change(params)
    assert folder in client.watched_folders

    # Test deletion
    params = {
        'folder': folder,
        'kind': WorkspaceUpdateKind.DELETION,
        'instance': ''
    }

    client.send_workspace_folders_change(params)
    assert folder not in client.watched_folders



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © 2018- Spyder Kernels Contributors
# Licensed under the terms of the MIT License
# (see spyder_kernels/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tests for the console kernel.
"""

# Standard library imports
import os

# Test imports
import pytest
from tornado import ioloop


# Local imports
from spyder_kernels.utils.test_utils import get_kernel
from spyder_kernels.comms.frontendcomm import FrontendComm
from spyder.plugins.ipythonconsole.comms.kernelcomm import KernelComm


# =============================================================================
# Fixtures
# =============================================================================
@pytest.fixture
def kernel(request):
    """Console kernel fixture"""
    # Get kernel instance
    kernel = get_kernel()
    kernel.io_loop = ioloop.IOLoop.current()
    kernel.namespace_view_settings = {
        'check_all': False,
        'exclude_private': True,
        'exclude_uppercase': True,
        'exclude_capitalized': False,
        'exclude_unsupported': False,
        'exclude_callables_and_modules': True,
        'excluded_names': [
            'nan',
            'inf',
            'infty',
            'little_endian',
            'colorbar_doc',
            'typecodes',
            '__builtins__',
            '__main__',
            '__doc__',
            'NaN',
            'Inf',
            'Infinity',
            'sctypes',
            'rcParams',
            'rcParamsDefault',
            'sctypeNA',
            'typeNA',
            'False_',
            'True_'
        ],
        'minmax': False,
        'filter_on': True}

    # Teardown
    def reset_kernel():
        kernel.do_execute('reset -f', True)
    request.addfinalizer(reset_kernel)
    return kernel


class dummyComm():
    def __init__(self):
        self.other = None
        self.message_callback = None
        self.close_callback = None
        self.comm_id = 1

    def close(self):
        self.other.close_callback({'content': {'comm_id': self.comm_id}})

    def send(self, msg_dict, buffers=None):
        msg = {
            'buffers': buffers,
            'content': {'data': msg_dict, 'comm_id': self.comm_id},
            }
        self.other.message_callback(msg)

    def _send_msg(self, *args, **kwargs):
        pass

    def on_msg(self, callback):
        self.message_callback = callback

    def on_close(self, callback):
        self.close_callback = callback


@pytest.fixture
def comms(kernel):
    """Get the comms"""
    commA = dummyComm()
    commB = dummyComm()
    commA.other = commB
    commB.other = commA

    frontend_comm = FrontendComm(kernel)
    kernel_comm = KernelComm()

    class DummyKernelClient():
        shell_channel = 0
        control_channel = 0

        def is_alive():
            return True

    kernel_comm.kernel_client = DummyKernelClient()

    kernel_comm._register_comm(commA)

    # Bypass the target system as this is not what is being tested
    frontend_comm._comm_open(commB, {'content': {}})

    return (kernel_comm, frontend_comm)


# =============================================================================
# Tests
# =============================================================================
@pytest.mark.skipif(os.name == 'nt', reason="Hangs on Windows")
def test_comm_base(comms):
    """Test basic message exchange."""
    commsend, commrecv = comms

    assert commsend.is_open()
    assert commrecv.is_open()

    received_messages = []

    def handler(msg_dict, buffers):
        received_messages.append((msg_dict,))

    # Register callback
    commrecv._register_message_handler('test_message', handler)

    # Send a message
    commsend._send_message('test_message', content='content')
    assert len(received_messages) == 1
    assert received_messages[0][0]['spyder_msg_type'] == 'test_message'
    assert received_messages[0][0]['content'] == 'content'

    # Send another message
    commsend._send_message('test_message', content='content')
    assert len(received_messages) == 2

    # Unregister callback
    commrecv._register_message_handler('test_message', None)

    # Send another unhandled message
    commsend._send_message('test_message', content='content')
    assert len(received_messages) == 2


@pytest.mark.skipif(os.name == 'nt', reason="Hangs on Windows")
def test_request(comms):
    """Test if the requests are being replied to."""
    kernel_comm, frontend_comm = comms

    def handler(a, b):
        return a + b

    kernel_comm.register_call_handler('test_request', handler)

    res = frontend_comm.remote_call(blocking=True).test_request('a', b='b')

    assert res == 'ab'


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Working Directory Plugin.
"""

# Standard library imports
import os.path as osp

# Third party imports
from qtpy.QtCore import Signal

# Local imports
from spyder.api.plugins import SpyderPluginV2, Plugins
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.translations import _
from spyder.config.base import get_conf_path
from spyder.plugins.workingdirectory.confpage import WorkingDirectoryConfigPage
from spyder.plugins.workingdirectory.container import (
    WorkingDirectoryContainer)
from spyder.plugins.toolbar.api import ApplicationToolbars
from spyder.utils import encoding


class WorkingDirectory(SpyderPluginV2):
    """
    Working directory changer plugin.
    """

    NAME = 'workingdir'
    REQUIRES = [Plugins.Preferences, Plugins.Console, Plugins.Toolbar]
    OPTIONAL = [Plugins.Editor, Plugins.Explorer, Plugins.IPythonConsole,
                Plugins.Find, Plugins.Projects, Plugins.Run]
    CONTAINER_CLASS = WorkingDirectoryContainer
    CONF_SECTION = NAME
    CONF_WIDGET_CLASS = WorkingDirectoryConfigPage
    CAN_BE_DISABLED = False
    CONF_FILE = False
    LOG_PATH = get_conf_path(CONF_SECTION)

    # --- Signals
    # ------------------------------------------------------------------------
    sig_current_directory_changed = Signal(str)
    """
    This signal is emitted when the current directory has changed.

    Parameters
    ----------
    new_working_directory: str
        The new new working directory path.
    """

    # --- SpyderPluginV2 API
    # ------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _('Working directory')

    @staticmethod
    def get_description():
        return _("Manage the current working directory used in Spyder.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon('DirOpenIcon')

    def on_initialize(self):
        container = self.get_container()

        container.sig_current_directory_changed.connect(
            self.sig_current_directory_changed)
        self.sig_current_directory_changed.connect(
            lambda path, plugin=None: self.chdir(path, plugin))

        cli_options = self.get_command_line_options()
        container.set_history(
            self.load_history(),
            cli_options.working_directory
        )

    @on_plugin_available(plugin=Plugins.Toolbar)
    def on_toolbar_available(self):
        container = self.get_container()
        toolbar = self.get_plugin(Plugins.Toolbar)
        toolbar.add_application_toolbar(container.toolbar)

    @on_plugin_available(plugin=Plugins.Preferences)
    def on_preferences_available(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(self)

    @on_plugin_available(plugin=Plugins.Editor)
    def on_editor_available(self):
        editor = self.get_plugin(Plugins.Editor)
        editor.sig_dir_opened.connect(self._editor_change_dir)
        container = self.get_container()
        container.edit_goto.connect(editor.load)

    @on_plugin_available(plugin=Plugins.Explorer)
    def on_explorer_available(self):
        explorer = self.get_plugin(Plugins.Explorer)
        self.sig_current_directory_changed.connect(self._explorer_change_dir)
        explorer.sig_dir_opened.connect(self._explorer_dir_opened)

    @on_plugin_available(plugin=Plugins.IPythonConsole)
    def on_ipyconsole_available(self):
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)

        self.sig_current_directory_changed.connect(
            ipyconsole.set_current_client_working_directory)
        ipyconsole.sig_current_directory_changed.connect(
            self._ipyconsole_change_dir)

    @on_plugin_available(plugin=Plugins.Projects)
    def on_projects_available(self):
        projects = self.get_plugin(Plugins.Projects)
        projects.sig_project_loaded.connect(self._project_loaded)
        projects.sig_project_closed[str].connect(self._project_closed)

    @on_plugin_teardown(plugin=Plugins.Toolbar)
    def on_toolbar_teardown(self):
        toolbar = self.get_plugin(Plugins.Toolbar)
        toolbar.remove_application_toolbar(
            ApplicationToolbars.WorkingDirectory)

    @on_plugin_teardown(plugin=Plugins.Preferences)
    def on_preferences_teardown(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.deregister_plugin_preferences(self)

    @on_plugin_teardown(plugin=Plugins.Editor)
    def on_editor_teardown(self):
        editor = self.get_plugin(Plugins.Editor)
        editor.sig_dir_opened.disconnect(self._editor_change_dir)

    @on_plugin_teardown(plugin=Plugins.Explorer)
    def on_explorer_teardown(self):
        explorer = self.get_plugin(Plugins.Explorer)
        self.sig_current_directory_changed.disconnect(self._explorer_change_dir)
        explorer.sig_dir_opened.disconnect(self._explorer_dir_opened)

    @on_plugin_teardown(plugin=Plugins.IPythonConsole)
    def on_ipyconsole_teardown(self):
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)

        self.sig_current_directory_changed.disconnect(
            ipyconsole.set_current_client_working_directory)
        ipyconsole.sig_current_directory_changed.disconnect(
            self._ipyconsole_change_dir)

    @on_plugin_teardown(plugin=Plugins.Projects)
    def on_projects_teardown(self):
        projects = self.get_plugin(Plugins.Projects)
        projects.sig_project_loaded.disconnect(self._project_loaded)
        projects.sig_project_closed[str].disconnect(self._project_closed)

    # --- Public API
    # ------------------------------------------------------------------------
    def chdir(self, directory, sender_plugin=None):
        """
        Change current working directory.

        Parameters
        ----------
        directory: str
            The new working directory to set.
        sender_plugin: spyder.api.plugins.SpyderPluginsV2
            The plugin that requested this change: Default is None.
        """
        explorer = self.get_plugin(Plugins.Explorer)
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)
        find = self.get_plugin(Plugins.Find)
        run = self.get_plugin(Plugins.Run)

        if explorer and sender_plugin != explorer:
            explorer.chdir(directory, emit=False)
            explorer.refresh(directory, force_current=True)

        if ipyconsole and sender_plugin != ipyconsole:
            ipyconsole.set_current_client_working_directory(directory)

        if find:
            find.refresh_search_directory()

        # This is a quick hack to make the Run plugin use the current working
        # directory when the option for it is set by users. In 6.1 we'll
        # improve how other plugins receive changes to the cwd from this one
        # to avoid things like this.
        # Fixes spyder-ide/spyder#23866
        if run:
            run._switch_working_dir(directory)

        if sender_plugin is not None:
            container = self.get_container()
            container.chdir(directory, emit=False)
            if ipyconsole:
                ipyconsole.save_working_directory(directory)

        self.save_history()

    def load_history(self, workdir=None):
        """
        Load history from a text file located in Spyder configuration folder
        or use `workdir` if there are no directories saved yet.

        Parameters
        ----------
        workdir: str
            The working directory to return. Default is None.
        """
        if osp.isfile(self.LOG_PATH):
            history, _ = encoding.readlines(self.LOG_PATH)
            history = [name for name in history if osp.isdir(name)]
        else:
            if workdir is None:
                workdir = self.get_container()._get_init_workdir()

            history = [workdir]

        return history

    def save_history(self):
        """
        Save history to a text file located in Spyder configuration folder.
        """
        history = self.get_container().get_history()
        try:
            encoding.writelines(history, self.LOG_PATH)
        except EnvironmentError:
            pass

    def get_workdir(self):
        """
        Get current working directory.

        Returns
        -------
        str
            Current working directory.
        """
        return self.get_container().get_workdir()

    # -------------------------- Private API ----------------------------------
    def _editor_change_dir(self, path):
        editor = self.get_plugin(Plugins.Editor)
        self.chdir(path, editor)

    def _explorer_change_dir(self, path):
        explorer = self.get_plugin(Plugins.Explorer)
        if explorer:
            explorer.chdir(path, emit=False)

    def _explorer_dir_opened(self, path):
        explorer = self.get_plugin(Plugins.Explorer)
        self.chdir(path, explorer)

    def _ipyconsole_change_dir(self, path):
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)
        self.chdir(path, ipyconsole)

    def _project_loaded(self, path):
        projects = self.get_plugin(Plugins.Projects)
        self.chdir(directory=path, sender_plugin=projects)

    def _project_closed(self, path):
        projects = self.get_plugin(Plugins.Projects)
        self.chdir(
            directory=projects.get_last_working_dir(),
            sender_plugin=projects
        )



# -*- coding: utf-8 -*-
"""
Spyder Editor

This is a temporary script file.
"""

import os
import sys


# %% functions
def d():
    def inner():
        return 2

    return inner


# ---- func 1 and 2
def func1():
    for i in range(
        3
    ):
        print(i)


def func2():
    if True:
        pass


# ---- other functions
def a():
    pass


def b():
    pass


def c():
    pass


# %% classes
class Class1:
    def __init__(
        self,
    ):
        super(
            Class1,
            self,
        ).__init__()
        self.x = 2

    def method3(
        self,
    ):
        pass

    def method2(
        self,
    ):
        pass

    def method1(
        self,
    ):
        pass



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License

"""Tests for encodings.py"""

import os
import pathlib
import stat

from flaky import flaky
import pytest

from spyder.utils.encoding import is_text_file, read, write
from spyder.py3compat import to_text_string

__location__ = os.path.realpath(os.path.join(os.getcwd(),
                                             os.path.dirname(__file__)))


@pytest.mark.order(1)
def test_symlinks(tmpdir):
    """
    Check that modifying symlinks files changes source file and keeps symlinks.
    """
    base_dir = tmpdir.mkdir("symlinks")
    base_file = base_dir.join("symlinks_text.txt")
    base_file_path = to_text_string(base_file)

    # Write base file
    write("Some text for symlink", base_file_path)

    # Create symlink
    symlink_file = pathlib.Path(base_dir.join(
        'link-to-symlinks_text.txt'))
    symlink_file.symlink_to(base_file_path)
    symlink_file_path = to_text_string(symlink_file)

    # Assert the symlink was created
    assert os.path.islink(symlink_file_path)

    # Write using the symlink
    encoding = write("New text for symlink", symlink_file_path)

    # Assert symlink is valid and contents of the file
    assert os.path.islink(symlink_file_path)
    assert base_file.read_text(encoding) == symlink_file.read_text(encoding)
    assert symlink_file.read_text(encoding) == 'New text for symlink'


def test_permissions(tmpdir):
    """Check that file permissions are preserved."""
    p_file = tmpdir.mkdir("permissions").join("permissions_text.txt")
    p_file = to_text_string(p_file)

    # Write file and define execution permissions
    write("Some text", p_file)
    st = os.stat(p_file)
    mode = st.st_mode | stat.S_IEXEC
    os.chmod(p_file, mode)

    old_mode = os.stat(p_file).st_mode

    # Write the file and check permissions
    write("Some text and more", p_file)
    new_mode = os.stat(p_file).st_mode

    assert old_mode == new_mode


@flaky(max_runs=10)
def test_timestamp(tmpdir):
    """Check that the modification timestamp is preserved."""
    tmp_file = tmpdir.mkdir("timestamp").join('test_file.txt')
    tmp_file = to_text_string(tmp_file)

    # Write a file
    write("Test text", tmp_file)
    st = os.stat(tmp_file)
    actual_creation_time = st.st_atime

    # Write the file and check that creation time is preserved.
    write('New text', tmp_file)
    creation_time = os.stat(tmp_file).st_atime
    assert actual_creation_time == creation_time


def test_is_text_file(tmpdir):
    p = tmpdir.mkdir("sub").join("random_text.txt")
    p.write("Some random text")
    assert is_text_file(str(p)) is True


@pytest.mark.parametrize(
    'expected_encoding, text_file',
    [('utf-8', 'utf-8.txt'),
     ('windows-1252', 'windows-1252.txt'),
     ('ascii', 'ascii.txt'),
     ('Big5', 'Big5.txt'),
     ('KOI8-R', 'KOI8-R.txt'),
     ('iso-8859-1', 'copyright.txt'),
     ('utf-8', 'copyright.py'),  # Python files are UTF-8 by default
     ('iso8859-9', 'iso8859-9.py')  # Encoding declared in file
     ])
def test_files_encodings(expected_encoding, text_file):
    file_path = os.path.join(__location__, text_file)
    text, encoding = read(file_path)
    assert encoding.lower() == expected_encoding.lower()


if __name__ == '__main__':
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Spyder API menu widgets.
"""

# Standard library imports
import sys
from typing import Optional, Union, TypeVar

# Third party imports
import qstylizer.style
from qtpy.QtCore import QTimer
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import QAction, QMenu, QProxyStyle, QStyle, QWidget

# Local imports
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.utils.qthelpers import add_actions, set_menu_icons, SpyderAction
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle, MAC, WIN


# ---- Constants
# -----------------------------------------------------------------------------
MENU_SEPARATOR = None


# Generic type annotations
T = TypeVar('T', bound='SpyderMenu')


class OptionsMenuSections:
    Top = 'top_section'
    Bottom = 'bottom_section'


class PluginMainWidgetMenus:
    Context = 'context_menu'
    Options = 'options_menu'


# ---- Style
# -----------------------------------------------------------------------------
class SpyderMenuProxyStyle(QProxyStyle):
    """Style adjustments that can only be done with a proxy style."""

    def pixelMetric(self, metric, option=None, widget=None):
        if metric == QStyle.PM_SmallIconSize:
            # Change icon size for menus.
            # Taken from https://stackoverflow.com/a/42145885/438386
            delta = -1 if MAC else (0 if WIN else 1)

            return (
                QProxyStyle.pixelMetric(self, metric, option, widget) + delta
            )

        return QProxyStyle.pixelMetric(self, metric, option, widget)


# ---- Widgets
# -----------------------------------------------------------------------------
class SpyderMenu(QMenu, SpyderFontsMixin):
    """
    A QMenu subclass to implement additional functionality for Spyder.
    """
    MENUS = []
    APP_MENU = False
    HORIZONTAL_MARGIN_FOR_ITEMS = 2 * AppStyle.MarginSize
    HORIZONTAL_PADDING_FOR_ITEMS = 3 * AppStyle.MarginSize

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        menu_id: Optional[str] = None,
        title: Optional[str] = None,
        min_width: Optional[int] = None,
        reposition: Optional[bool] = True,
    ):
        """
        Create a menu for Spyder.

        Parameters
        ----------
        parent: QWidget or None
            The menu's parent
        menu_id: str
            Unique str identifier for the menu.
        title: str or None
            Localized text string for the menu.
        min_width: int or None
            Minimum width for the menu.
        reposition: bool, optional (default True)
            Whether to vertically reposition the menu due to it's padding.
        """
        self._parent = parent
        self.menu_id = menu_id
        self._title = title
        self._reposition = reposition

        self._sections = []
        self._actions = []
        self._actions_map = {}
        self._unintroduced_actions = {}
        self._after_sections = {}
        self._dirty = False
        self._is_shown = False
        self._is_submenu = False
        self._in_app_menu = False

        if title is None:
            super().__init__(parent)
        else:
            super().__init__(title, parent)

        self.MENUS.append((parent, title, self))

        # Set min width
        if min_width is not None:
            self.setMinimumWidth(min_width)

        # Signals
        self.aboutToShow.connect(self.render)

        # Adjustmens for Mac
        if sys.platform == 'darwin':
            # Needed to enable the dynamic population of actions in app menus
            # in the aboutToShow signal.
            # See spyder-ide/spyder#14612
            if self.APP_MENU:
                self.addAction(QAction(self))

            # Necessary to follow Mac's HIG for app menus.
            self.aboutToShow.connect(self._set_icons)

        # Style
        self.css = self._generate_stylesheet()
        self.setStyleSheet(self.css.toString())

        style = SpyderMenuProxyStyle(None)
        style.setParent(self)
        self.setStyle(style)

    # ---- Public API
    # -------------------------------------------------------------------------
    def clear_actions(self):
        """
        Remove actions from the menu (including custom references)

        Returns
        -------
        None.
        """
        self.clear()
        self._sections = []
        self._actions = []
        self._actions_map = {}
        self._unintroduced_actions = {}
        self._after_sections = {}

    def add_action(self: T,
                   action: Union[SpyderAction, T],
                   section: Optional[str] = None,
                   before: Optional[str] = None,
                   before_section: Optional[str] = None,
                   check_before: bool = True,
                   omit_id: bool = False):
        """
        Add action to a given menu section.

        Parameters
        ----------
        action: SpyderAction
            The action to add.
        section: str or None
            The section id in which to insert the `action`.
        before: str
            Make the action appear before the given action identifier.
        before_section: str or None
            Make the item section (if provided) appear before another
            given section.
        check_before: bool
            Check if the `before` action is part of the menu. This is
            necessary to avoid an infinite recursion when adding
            unintroduced actions with this method again.
        omit_id: bool
            If True, then the menu will check if the item to add declares an
            id, False otherwise. This flag exists only for items added on
            Spyder 4 plugins. Default: False
        """
        item_id = None
        if isinstance(action, SpyderAction) or hasattr(action, 'action_id'):
            item_id = action.action_id

            # This is necessary when we set a menu for `action`, e.g. for
            # todo_list_action in EditorMainWidget.
            if action.menu() and isinstance(action.menu(), SpyderMenu):
                action.menu()._is_submenu = True
        elif isinstance(action, SpyderMenu) or hasattr(action, 'menu_id'):
            item_id = action.menu_id
            action._is_submenu = True

        if not omit_id and item_id is None and action is not None:
            raise AttributeError(f'Item {action} must declare an id.')

        if before is None:
            self._actions.append((section, action))
        else:
            new_actions = []
            added = False
            before_item = self._actions_map.get(before, None)

            for sec, act in self._actions:
                if before_item is not None and act == before_item:
                    added = True
                    new_actions.append((section, action))

                new_actions.append((sec, act))

            # Actions can't be added to the menu if the `before` action is
            # not part of it yet. That's why we need to save them in the
            # `_unintroduced_actions` dict, so we can add them again when
            # the menu is rendered.
            if not added and check_before:
                before_actions = self._unintroduced_actions.get(before, [])
                before_actions.append((section, action))
                self._unintroduced_actions[before] = before_actions

            self._actions = new_actions

        if section not in self._sections:
            self._add_section(section, before_section)

        # Track state of menu to avoid re-rendering if menu has not changed
        self._dirty = True
        self._actions_map[item_id] = action

    def remove_action(self, item_id: str):
        if item_id in self._actions_map:
            action = self._actions_map.pop(item_id)
            position = None

            for i, (_, act) in enumerate(self._actions):
                if act == action:
                    position = i
                    break

            if position is not None:
                self._actions.pop(position)
                self._dirty = True

    def get_title(self):
        """
        Return the title for menu.
        """
        return self._title

    def get_actions(self):
        """
        Return a parsed list of menu actions.

        Includes MENU_SEPARATOR taking into account the sections defined.
        """
        actions = []
        for section in self._sections:
            for (sec, action) in self._actions:
                if sec == section:
                    actions.append(action)

            actions.append(MENU_SEPARATOR)
        return actions

    def get_sections(self):
        """
        Return a tuple of menu sections.
        """
        return tuple(self._sections)

    # ---- Private API
    # -------------------------------------------------------------------------
    def _add_missing_actions(self):
        """
        Add actions that were not introduced to the menu because a `before`
        action they require is not part of it.
        """
        for before, actions in self._unintroduced_actions.items():
            for section, action in actions:
                self.add_action(
                    action,
                    section=section,
                    before=before,
                    check_before=False
                )

        self._unintroduced_actions = {}

    def render(self, force=False):
        """
        Create the menu prior to showing it. This takes into account sections
        and location of menus.

        Parameters
        ----------
        force: bool, optional
            Whether to force rendering the menu.
        """
        if self._dirty or force:
            self.clear()
            self._add_missing_actions()

            actions = self.get_actions()
            add_actions(self, actions)
            self._set_icons()

            self._dirty = False

    def _add_section(self, section, before_section=None):
        """
        Add a new section to the list of sections in this menu.

        Parameters
        ----------
        before_section: str or None
            Make `section` appear before another one.
        """
        inserted_before_other = False

        if before_section is not None:
            if before_section in self._sections:
                # If before_section was already introduced, we simply need to
                # insert the new section on its position, which will put it
                # exactly behind before_section.
                idx = self._sections.index(before_section)
                self._sections.insert(idx, section)
                inserted_before_other = True
            else:
                # If before_section hasn't been introduced yet, we know we need
                # to insert it after section when it's finally added to the
                # menu. So, we preserve that info in the _after_sections dict.
                self._after_sections[before_section] = section

                # Append section to the list of sections because we assume
                # people build menus from top to bottom, i.e. they add its
                # upper sections first.
                self._sections.append(section)
        else:
            self._sections.append(section)

        # Check if section should be inserted after another one, according to
        # what we have in _after_sections.
        after_section = self._after_sections.pop(section, None)

        if after_section is not None:
            if not inserted_before_other:
                # Insert section to the right of after_section, if it was not
                # inserted before another one.
                if section in self._sections:
                    self._sections.remove(section)

                index = self._sections.index(after_section)
                self._sections.insert(index + 1, section)
            else:
                # If section was already inserted before another one, then we
                # need to move after_section to its left instead.
                if after_section in self._sections:
                    self._sections.remove(after_section)

                idx = self._sections.index(section)
                idx = idx if (idx == 0) else (idx - 1)
                self._sections.insert(idx, after_section)

    def _set_icons(self):
        """
        Unset menu icons for app menus and set them for regular menus.

        This is necessary only for Mac to follow its Human Interface
        Guidelines (HIG), which don't recommend icons in app menus.
        """
        if sys.platform == "darwin":
            if self.APP_MENU or self._in_app_menu:
                set_menu_icons(self, False, in_app_menu=True)
            else:
                set_menu_icons(self, True)

    @classmethod
    def _generate_stylesheet(cls):
        """Generate base stylesheet for menus."""
        css = qstylizer.style.StyleSheet()
        font = cls.get_font(SpyderFontType.Interface)

        # Add padding and border to follow modern standards
        css.QMenu.setValues(
            # Only add top and bottom padding so that menu separators can go
            # completely from the left to right border.
            paddingTop=f'{2 * AppStyle.MarginSize}px',
            paddingBottom=f'{2 * AppStyle.MarginSize}px',
            # This uses the same color as the separator
            border=f"1px solid {SpyderPalette.COLOR_BACKGROUND_6}"
        )

        # Set the right background color. This is the only way to do it!
        css['QWidget:disabled QMenu'].setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_3,
        )

        # Add padding around separators to prevent that hovering on items hides
        # them.
        css["QMenu::separator"].setValues(
            # Only add top and bottom margins so that the separators can go
            # completely from the left to right border.
            margin=f'{2 * AppStyle.MarginSize}px 0px',
        )

        # Set menu item properties
        delta_top = 0 if (MAC or WIN) else 1
        delta_bottom = 0 if MAC else (2 if WIN else 1)
        css["QMenu::item"].setValues(
            height='1.1em' if MAC else ('1.35em' if WIN else '1.25em'),
            marginLeft=f'{cls.HORIZONTAL_MARGIN_FOR_ITEMS}px',
            marginRight=f'{cls.HORIZONTAL_MARGIN_FOR_ITEMS}px',
            paddingTop=f'{AppStyle.MarginSize + delta_top}px',
            paddingBottom=f'{AppStyle.MarginSize + delta_bottom}px',
            paddingLeft=f'{cls.HORIZONTAL_PADDING_FOR_ITEMS}px',
            paddingRight=f'{cls.HORIZONTAL_PADDING_FOR_ITEMS}px',
            fontFamily=font.family(),
            fontSize=f'{font.pointSize()}pt',
            backgroundColor='transparent'
        )

        # Set hover and pressed state of items
        for state in ['selected', 'pressed']:
            if state == 'selected':
                bg_color = SpyderPalette.COLOR_BACKGROUND_4
            else:
                bg_color = SpyderPalette.COLOR_BACKGROUND_5

            css[f"QMenu::item:{state}"].setValues(
                backgroundColor=bg_color,
                borderRadius=SpyderPalette.SIZE_BORDER_RADIUS
            )

        # Set disabled state of items
        for state in ['disabled', 'selected:disabled']:
            css[f"QMenu::item:{state}"].setValues(
                color=SpyderPalette.COLOR_DISABLED,
                backgroundColor="transparent"
            )

        return css

    def __str__(self):
        return f"SpyderMenu('{self.menu_id}')"

    def __repr__(self):
        return f"SpyderMenu('{self.menu_id}')"

    def _adjust_menu_position(self):
        """Menu position adjustment logic to follow custom style."""
        if not self._is_shown:
            # Reposition submenus vertically due to padding and border
            if self._reposition and self._is_submenu:
                self.move(
                    self.pos().x(),
                    # Current vertical pos - padding - border
                    self.pos().y() - 2 * AppStyle.MarginSize - 1
                )

            self._is_shown = True

        # Reposition menus horizontally due to border
        if self.APP_MENU:
            delta_x = 0 if MAC else 3
        else:
            if QCursor().pos().x() - self.pos().x() < 40:
                # If the difference between the current cursor x position and
                # the menu one is small, it means the menu will be shown to the
                # right, so we need to move it in that direction.
                delta_x = 1
            else:
                # This happens when the menu is shown to the left.
                delta_x = -1

        self.move(self.pos().x() + delta_x, self.pos().y())

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def showEvent(self, event):
        """Call adjustments when the menu is going to be shown."""
        # To prevent race conditions which can cause partially showing a menu
        # (as in spyder-ide/spyder#22266), we use a timer to queue the move
        # related events after the menu is shown.
        # For more info you can check:
        #  * https://forum.qt.io/topic/23381/showevent-not-working/3
        #  * https://stackoverflow.com/a/49351518
        QTimer.singleShot(0, self._adjust_menu_position)


class PluginMainWidgetOptionsMenu(SpyderMenu):
    """
    Options menu for PluginMainWidget.
    """

    def render(self):
        """Render the menu's bottom section as expected."""
        if self._dirty:
            self.clear()
            self._add_missing_actions()

            bottom = OptionsMenuSections.Bottom
            actions = []
            for section in self._sections:
                for (sec, action) in self._actions:
                    if sec == section and sec != bottom:
                        actions.append(action)

                actions.append(MENU_SEPARATOR)

            # Add bottom actions
            for (sec, action) in self._actions:
                if sec == bottom:
                    actions.append(action)

            add_actions(self, actions)
            self._set_icons()

            self._dirty = False



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Standard library imports
import os.path as osp

# Third party imports
from qtpy.compat import getexistingdirectory
from qtpy.QtCore import QEvent, Qt, Signal, Slot
from qtpy.QtWidgets import QMessageBox, QSizePolicy

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.comboboxes import SpyderComboBox
from spyder.utils.encoding import to_unicode_from_fs


# ---- Constants
# ----------------------------------------------------------------------------
class SearchInComboBoxItems:
    Cwd = 0
    Project = 1
    File = 2
    FirstSeparator = 3
    SelectAnotherDirectory = 4
    ClearList = 5
    SecondSeparator = 6
    ExternalPaths = 7

MAX_PATH_HISTORY = 15


# ---- Combobox
# ----------------------------------------------------------------------------
class SearchInComboBox(SpyderComboBox):
    """
    Non editable combo box handling the path locations of the FindOptions
    widget.
    """

    # Signals
    sig_redirect_stdio_requested = Signal(bool)

    def __init__(self, external_path_history=[], parent=None, id_=None):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.setEditable(False)

        self.path = ''
        self.project_path = ''
        self.file_path = ''
        self.external_path = ''

        if id_ is not None:
            self.ID = id_

        self.addItem(_("Current working directory"))

        self.addItem(_("Project"))
        self.model().item(SearchInComboBoxItems.Project, 0).setEnabled(False)

        self.addItem(_("Current file").replace('&', ''))

        self.insertSeparator(SearchInComboBoxItems.FirstSeparator)

        self.addItem(_("Select another directory"))
        self.addItem(_("Clear the list of other directories"))

        self.insertSeparator(SearchInComboBoxItems.SecondSeparator)

        if external_path_history:
            for path in external_path_history:
                self.add_external_path(path)
        else:
            self.set_state_other_dirs_items(False)

        self.currentIndexChanged.connect(self.path_selection_changed)
        self.view().installEventFilter(self)

    def add_external_path(self, path):
        """
        Adds an external path to the combobox if it exists on the file system.
        If the path is already listed in the combobox, it is removed from its
        current position and added back at the end. If the maximum number of
        paths is reached, the oldest external path is removed from the list.
        """
        if not osp.exists(path):
            return
        self.set_state_other_dirs_items(True)
        self.removeItem(self.findText(path))
        self.addItem(path)
        self.setItemData(self.count() - 1, path, Qt.ToolTipRole)

        while (
            self.count() >
            (MAX_PATH_HISTORY + SearchInComboBoxItems.ExternalPaths)
        ):
            self.removeItem(SearchInComboBoxItems.ExternalPaths)

    def get_external_paths(self):
        """Returns a list of the external paths listed in the combobox."""
        return [
            str(self.itemText(i))
            for i in range(SearchInComboBoxItems.ExternalPaths, self.count())
        ]

    def clear_external_paths(self):
        """Remove all the external paths listed in the combobox."""
        while self.count() > SearchInComboBoxItems.ExternalPaths:
            self.removeItem(SearchInComboBoxItems.ExternalPaths)
        self.set_state_other_dirs_items(False)

    def get_current_searchpath(self):
        """
        Returns the path corresponding to the currently selected item
        in the combobox.
        """
        idx = self.currentIndex()
        if idx == SearchInComboBoxItems.Cwd:
            return self.path
        elif idx == SearchInComboBoxItems.Project:
            return self.project_path
        elif idx == SearchInComboBoxItems.File:
            return self.file_path
        else:
            return self.external_path

    def set_current_searchpath_index(self, index):
        """Set the current index of this combo box."""
        if index is not None:
            index = min(index, self.count() - 1)
            if index in [SearchInComboBoxItems.ClearList,
                         SearchInComboBoxItems.SelectAnotherDirectory]:
                index = SearchInComboBoxItems.Cwd
        else:
            index = SearchInComboBoxItems.Cwd

        self.setCurrentIndex(index)

    def is_file_search(self):
        """Returns whether the current search path is a file."""
        if self.currentIndex() == SearchInComboBoxItems.File:
            return True
        else:
            return False

    @Slot()
    def path_selection_changed(self):
        """Handles when the current index of the combobox changes."""
        idx = self.currentIndex()
        if idx == SearchInComboBoxItems.SelectAnotherDirectory:
            external_path = self.select_directory()
            if len(external_path) > 0:
                self.add_external_path(external_path)
                self.setCurrentIndex(self.count() - 1)
            else:
                self.setCurrentIndex(SearchInComboBoxItems.Cwd)
        elif idx == SearchInComboBoxItems.ClearList:
            reply = QMessageBox.question(
                    self, _("Clear other directories"),
                    _("Do you want to clear the list of other directories?"),
                    QMessageBox.Yes | QMessageBox.No)
            if reply == QMessageBox.Yes:
                self.clear_external_paths()
            self.setCurrentIndex(SearchInComboBoxItems.Cwd)
        elif idx >= SearchInComboBoxItems.ExternalPaths:
            self.external_path = str(self.itemText(idx))

    @Slot()
    def select_directory(self):
        """Select directory"""
        self.sig_redirect_stdio_requested.emit(False)
        directory = getexistingdirectory(
            self,
            _("Select directory"),
            self.path,
        )
        if directory:
            directory = to_unicode_from_fs(osp.abspath(directory))

        self.sig_redirect_stdio_requested.emit(True)
        return directory

    def set_project_path(self, path):
        """
        Sets the project path and disables the project search in the combobox
        if the value of path is None.
        """
        if path is None:
            self.project_path = ''
            self.model().item(
                SearchInComboBoxItems.Project, 0
            ).setEnabled(False)
            if self.currentIndex() == SearchInComboBoxItems.Project:
                self.setCurrentIndex(SearchInComboBoxItems.Cwd)
        else:
            path = osp.abspath(path)
            self.project_path = path
            self.model().item(
                SearchInComboBoxItems.Project, 0
            ).setEnabled(True)

    def set_state_other_dirs_items(self, enabled):
        """
        Set the enabled/visible state of items that change when other
        directories are added/removed to/from the combobox.
        """
        # The second separator needs to be visible only when the user has added
        # other directories.
        self.view().setRowHidden(
            SearchInComboBoxItems.SecondSeparator, not enabled
        )

        # The ClearList item needs to be disabled if the user has not added
        # other directories
        self.model().item(
            SearchInComboBoxItems.ClearList, 0
        ).setEnabled(enabled)

    def eventFilter(self, widget, event):
        """Used to handle key events on the QListView of the combobox."""
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Delete:
            index = self.view().currentIndex().row()
            if index >= SearchInComboBoxItems.ExternalPaths:
                # Remove item and update the view.
                self.removeItem(index)
                self.showPopup()

                # Set the view selection so that it doesn't bounce around.
                new_index = min(self.count() - 1, index)
                if new_index < SearchInComboBoxItems.ExternalPaths:
                    new_index = SearchInComboBoxItems.Cwd
                    self.set_state_other_dirs_items(False)
                    self.hidePopup()

                self.view().setCurrentIndex(self.model().index(new_index, 0))
                self.setCurrentIndex(new_index)

            return True

        return super().eventFilter(widget, event)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Testing utilities for the CodeEditor to be used with pytest.
"""

# Standard library imports
import os
import os.path as osp
from unittest.mock import Mock

# Third party imports
import pytest
from qtpy.QtGui import QFont
from qtpy.QtCore import QMimeData, QUrl
from qtpy.QtWidgets import QApplication

# Local imports
from spyder.config.manager import CONF
from spyder.plugins.completion.tests.conftest import (
    completion_plugin_all,
    completion_plugin_all_started,
    MainWindowMock,
    qtbot_module
)
from spyder.plugins.editor.widgets.codeeditor import CodeEditor
from spyder.plugins.explorer.widgets.tests.conftest import create_folders_files
from spyder.plugins.outlineexplorer.editor import OutlineExplorerProxyEditor
from spyder.plugins.outlineexplorer.main_widget import OutlineExplorerWidget
from spyder.utils.programs import is_module_installed
from spyder.widgets.findreplace import FindReplace


# ---- Base constants for assets
HERE = osp.dirname(osp.abspath(__file__))
ASSETS = osp.join(HERE, 'assets')

# ---- Constants for outline related test
AVAILABLE_CASES = ['text']
CASES = {
    case: {
        'file': osp.join(ASSETS, '{0}.py'.format(case)),
        'tree': osp.join(ASSETS, '{0}_trees.json'.format(case))
    }
    for case in AVAILABLE_CASES
}

# ---- Auxiliary constants and functions for formatting tests
autopep8 = pytest.param(
    'autopep8',
    marks=pytest.mark.skipif(
        os.name == 'nt',
        reason='autopep8 produces a different output on Windows'
    )
)

yapf = pytest.param(
    'yapf',
    marks=pytest.mark.skipif(
        is_module_installed('yapf', '<0.32.0'),
        reason='Versions older than 0.32 produce different outputs'
    )
)

black = pytest.param(
    'black',
    marks=pytest.mark.skipif(
        is_module_installed('python-lsp-black', '<2.0.0'),
        reason='Versions older than 2.0 use a different entrypoint name'
    )
)


def get_formatter_values(formatter, newline, range_fmt=False, max_line=False):
    if range_fmt:
        suffix = 'range'
    elif max_line:
        suffix = 'max_line'
    else:
        suffix = 'result'

    original_file = osp.join(ASSETS, 'original_file.py')
    formatted_file = osp.join(ASSETS, '{0}_{1}.py'.format(formatter, suffix))

    with open(original_file, 'r') as f:
        text = f.read()
    text = text.replace('\n', newline)

    with open(formatted_file, 'r') as f:
        result = f.read()
    result = result.replace('\n', newline)

    return text, result


# ---- Fixtures for outline functionality
@pytest.fixture
def outlineexplorer(qtbot):
    """Set up an OutlineExplorerWidget."""
    outlineexplorer = OutlineExplorerWidget(None, None, None)
    outlineexplorer.set_conf('show_fullpath', False)
    outlineexplorer.set_conf('show_all_files', True)
    outlineexplorer.set_conf('group_cells', True)
    outlineexplorer.set_conf('show_comments', True)
    outlineexplorer.set_conf('sort_files_alphabetically', False)
    outlineexplorer.set_conf('display_variables', True)

    # Fix the size of the outline explorer to prevent an
    # 'Unable to set geometry ' warning if the test fails.
    outlineexplorer.setFixedSize(400, 350)

    qtbot.addWidget(outlineexplorer)
    outlineexplorer.show()

    return outlineexplorer


# ---- Fixtures for CodeEditor class
def codeeditor_factory():
    editor = CodeEditor(parent=None)
    editor.setup_editor(language='Python',
                        tab_mode=False,
                        markers=True,
                        close_quotes=True,
                        close_parentheses=True,
                        color_scheme='spyder/dark',
                        font=QFont("Monospace", 10),
                        automatic_completions=True,
                        automatic_completions_after_chars=1,
                        folding=False)
    editor.eol_chars = '\n'
    editor.resize(640, 480)
    return editor


@pytest.fixture
def mock_completions_codeeditor(qtbot_module, request):
    """CodeEditor instance with ability to mock the completions response.

    Returns a tuple of (editor, mock_response). Tests using this fixture should
    set `mock_response.side_effect = lambda lang, method, params: {}`.
    """
    # Create a CodeEditor instance
    editor = codeeditor_factory()
    qtbot_module.addWidget(editor)
    editor.show()

    mock_response = Mock()

    def perform_request(lang, method, params):
        resp = mock_response(lang, method, params)
        print("DEBUG {}".format(resp))
        if resp is not None:
            editor.handle_response(method, resp)
    editor.sig_perform_completion_request.connect(perform_request)

    editor.filename = 'test.py'
    editor.language = 'Python'
    editor.completions_available = True
    qtbot_module.wait(2000)

    def teardown():
        editor.hide()
        editor.completion_widget.hide()
    request.addfinalizer(teardown)

    return editor, mock_response


@pytest.fixture
def completions_codeeditor(completion_plugin_all_started, qtbot_module,
                           request, capsys, tmp_path):
    """CodeEditor instance with LSP services activated."""
    # Create a CodeEditor instance
    editor = codeeditor_factory()
    qtbot_module.addWidget(editor)

    completion_plugin, capabilities = completion_plugin_all_started
    completion_plugin.wait_for_ms = 2000

    CONF.set('completions', 'enable_code_snippets', False)
    completion_plugin.after_configuration_update([])
    CONF.notify_section_all_observers('completions')

    # Redirect editor LSP requests to lsp_manager
    editor.sig_perform_completion_request.connect(
        completion_plugin.send_request)

    file_path = tmp_path / 'test.py'
    file_path.write_text('')

    editor.filename = str(file_path)
    editor.language = 'Python'

    completion_plugin.register_file('python', str(file_path), editor)
    editor.register_completion_capabilities(capabilities)

    with qtbot_module.waitSignal(
            editor.completions_response_signal, timeout=30000):
        editor.start_completion_services()

    def teardown():
        editor.completion_widget.hide()
        editor.tooltip_widget.hide()
        editor.hide()

        # Capture stderr and assert there are no errors
        sys_stream = capsys.readouterr()
        sys_err = sys_stream.err

        assert sys_err == ''

    request.addfinalizer(teardown)

    editor.show()

    return editor, completion_plugin


@pytest.fixture
def search_codeeditor(completions_codeeditor, qtbot_module, request):
    code_editor, _ = completions_codeeditor
    find_replace = FindReplace(code_editor, enable_replace=True)
    find_replace.set_editor(code_editor)
    qtbot_module.addWidget(find_replace)

    def teardown():
        find_replace.hide()

    request.addfinalizer(teardown)

    return code_editor, find_replace


@pytest.fixture
def codeeditor(qtbot):
    widget = CodeEditor(None)
    widget.setup_editor(linenumbers=True,
                        markers=True,
                        tab_mode=False,
                        font=QFont("Courier New", 10),
                        show_blanks=True, color_scheme='spyder/dark',
                        scroll_past_end=True)
    widget.setup_editor(language='Python')
    widget.resize(640, 480)
    widget.show()

    yield widget
    widget.close()


@pytest.fixture
def completions_codeeditor_outline(completions_codeeditor, outlineexplorer):
    editor, _ = completions_codeeditor
    editor.oe_proxy = OutlineExplorerProxyEditor(editor, editor.filename)
    outlineexplorer.register_editor(editor.oe_proxy)
    outlineexplorer.set_current_editor(
        editor.oe_proxy, update=False, clear=False)
    return editor, outlineexplorer


# ---- Other fixtures
@pytest.fixture
def copy_files_clipboard(create_folders_files):
    """Fixture to copy files/folders into the clipboard"""
    file_paths = create_folders_files[0]
    file_content = QMimeData()
    file_content.setUrls([QUrl.fromLocalFile(fname) for fname in file_paths])
    cb = QApplication.clipboard()
    cb.setMimeData(file_content, mode=cb.Clipboard)
    return file_paths



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Standard library imports
from __future__ import annotations
from typing import List, Type, Union

# Third party imports
import qstylizer.style
from qtpy.QtCore import QSize, Qt, Signal
from qtpy.QtGui import QFontMetricsF, QIcon
from qtpy.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLayout,
    QListView,
    QListWidget,
    QListWidgetItem,
    QScrollArea,
    QStackedWidget,
    QVBoxLayout,
    QWidget
)
from superqt.utils import qdebounced, signals_blocked

# Local imports
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.utils.icon_manager import ima
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import (
    AppStyle,
    MAC,
    PREFERENCES_TABBAR_STYLESHEET,
    WIN
)


class PageScrollArea(QScrollArea):
    """Scroll area for preference pages."""

    def widget(self):
        """Return the page widget inside the scroll area."""
        return super().widget().page


class SidebarPage(QWidget):
    """Base class for pages used in SidebarDialog's"""

    # Signals
    show_this_page = Signal()

    # Constants
    MAX_WIDTH = 620
    MIN_HEIGHT = 550

    def __init__(self, parent):
        QWidget.__init__(self, parent)

        # Set dimensions
        self.setMaximumWidth(self.MAX_WIDTH)
        self.setMinimumHeight(self.MIN_HEIGHT)

    def initialize(self):
        """Initialize page."""
        self.setup_page()

    def get_name(self):
        """Return page name."""
        raise NotImplementedError

    def get_icon(self):
        """Return page icon."""
        QIcon()

    def setup_page(self):
        """Setup widget to be shown in the page."""
        raise NotImplementedError

    @staticmethod
    def create_icon(name):
        """Create an icon by name using Spyder's icon manager."""
        return ima.icon(name)

    def sizeHint(self):
        """Default page size."""
        return QSize(self.MAX_WIDTH, self.MIN_HEIGHT)


class SidebarDialog(QDialog, SpyderFontsMixin):
    """Sidebar dialog."""

    # Constants
    ITEMS_MARGIN = 2 * AppStyle.MarginSize
    ITEMS_PADDING = (
        AppStyle.MarginSize if (MAC or WIN) else 2 * AppStyle.MarginSize
    )
    CONTENTS_WIDTH = 230 if MAC else (200 if WIN else 240)
    ICON_SIZE = 20
    PAGES_MINIMUM_WIDTH = 600

    # To be set by childs
    TITLE = ""
    ICON = QIcon()
    MIN_WIDTH = 800
    MIN_HEIGHT = 600
    PAGE_CLASSES: List[Type[SidebarPage]] = []

    def __init__(self, parent=None):
        QDialog.__init__(self, parent)

        # ---- Attributes
        self.items_font = self.get_font(
            SpyderFontType.Interface, font_size_delta=1
        )
        self._is_shown = False
        self._separators = []
        self._active_pages = {}

        # ---- Size
        self.setMinimumWidth(self.MIN_WIDTH)
        self.setMinimumHeight(self.MIN_HEIGHT)

        # ---- Widgets
        self.pages_widget = QStackedWidget(self)
        self.contents_widget = QListWidget(self)
        buttons_box, buttons_layout = self.create_buttons()

        # Destroying the C++ object right after closing the dialog box,
        # otherwise it may be garbage-collected in another QThread
        # (e.g. the editor's analysis thread in Spyder), thus leading to
        # a segmentation fault on UNIX or an application crash on Windows
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowTitle(self.TITLE)
        self.setWindowIcon(self.ICON)

        # ---- Widgets setup
        self.pages_widget.setMinimumWidth(self.PAGES_MINIMUM_WIDTH)

        self.contents_widget.setMovement(QListView.Static)
        self.contents_widget.setSpacing(3)
        self.contents_widget.setCurrentRow(0)
        self.contents_widget.setIconSize(QSize(self.ICON_SIZE, self.ICON_SIZE))
        self.contents_widget.setFixedWidth(self.CONTENTS_WIDTH)

        # Don't show horizontal scrollbar because it doesn't look good. Instead
        # we show tooltips if the text doesn't fit in contents_widget width.
        self.contents_widget.setHorizontalScrollBarPolicy(
            Qt.ScrollBarAlwaysOff
        )

        # ---- Layout
        contents_and_pages_layout = QGridLayout()
        contents_and_pages_layout.addWidget(self.contents_widget, 0, 0)
        contents_and_pages_layout.addWidget(self.pages_widget, 0, 1)
        contents_and_pages_layout.setContentsMargins(0, 0, 0, 0)
        contents_and_pages_layout.setColumnStretch(0, 1)
        contents_and_pages_layout.setColumnStretch(1, 3)
        contents_and_pages_layout.setHorizontalSpacing(0)

        layout = QVBoxLayout()
        layout.addLayout(contents_and_pages_layout)
        layout.addSpacing(
            - (2 * AppStyle.MarginSize) if MAC else AppStyle.MarginSize
        )
        layout.addLayout(buttons_layout)

        self.setLayout(layout)

        # ---- Stylesheet
        self.setStyleSheet(self._main_stylesheet)

        self._contents_css = self._generate_contents_stylesheet()
        self.contents_widget.setStyleSheet(self._contents_css.toString())

        self.contents_widget.verticalScrollBar().setStyleSheet(
            self._contents_scrollbar_stylesheet
        )

        # ---- Signals and slots
        self.pages_widget.currentChanged.connect(self.current_page_changed)
        self.contents_widget.currentRowChanged.connect(
            self.pages_widget.setCurrentIndex)
        buttons_box.accepted.connect(self.accept)
        buttons_box.rejected.connect(self.reject)
        buttons_box.clicked.connect(self.button_clicked)

        # Add pages to the dialog
        self._add_pages()

        # Set index to the initial page
        if self.PAGE_CLASSES:
            self.set_current_index(0)

    # ---- Public API to be overridden by children
    # -------------------------------------------------------------------------
    def button_clicked(self, button):
        """Actions to perform after one of the dialog's buttons is clicked."""
        pass

    def current_page_changed(self, index):
        """Actions to perform after the current page in the dialog changes."""
        pass

    def create_buttons(self):
        """
        Create the buttons that will be displayed in the dialog.

        Override this method if you want different buttons in it.
        """
        bbox = SpyderDialogButtonBox(QDialogButtonBox.Ok)

        layout = QHBoxLayout()
        layout.addWidget(bbox)

        return bbox, layout

    # ---- Public API
    # -------------------------------------------------------------------------
    def get_current_index(self):
        """Return current page index"""
        return self.contents_widget.currentRow()

    def set_current_index(self, index):
        """Set current page index"""
        self.contents_widget.setCurrentRow(index)

    def get_item(self, index: int | None = None) -> QListWidgetItem:
        """Get item on the left panel corresponding to `index`."""
        if index is None:
            index = self.get_current_index()
        return self.contents_widget.item(index)

    def get_page(
        self, index: int | None = None
    ) -> Union[QWidget, SidebarPage] | None:
        """Return page widget on the right panel corresponding to `index`."""
        if index is None:
            page = self.pages_widget.currentWidget()
        else:
            page = self.pages_widget.widget(index)

        # Not all pages are config pages (e.g. separators have a simple QWidget
        # as their config page). So, we need to check for this.
        if page and hasattr(page, 'widget'):
            return page.widget()

    def hide_page(self, index=None):
        """Hide page corresponding to `index`."""
        if index is None:
            index = self.get_current_index()
        self._active_pages[index] = False

        # Hide entry from contents_widget
        self.contents_widget.item(index).setHidden(True)

        # If the current page is not the first one (index=0), then we set as
        # the new current one the first active page before it.
        if index > 1:
            # With this we move backwards from index-1 up to 0 until we find
            # an active page.
            for i in range(index - 1, -1, -1):
                if self._active_pages.get(i):
                    self.set_current_index(i)
                    break
        elif index == 1:
            # If the current page is the second one, we set the first one as
            # current.
            self.set_current_index(0)

    def add_separator(self):
        """Add a horizontal line to separate different sections."""
        # Solution taken from https://stackoverflow.com/a/24819554/438386
        item = QListWidgetItem(self.contents_widget)
        item.setFlags(Qt.NoItemFlags)

        size = (
            AppStyle.MarginSize * 3 if (MAC or WIN)
            else AppStyle.MarginSize * 5
        )
        item.setSizeHint(QSize(size, size))

        hline = QFrame(self.contents_widget)
        hline.setFrameShape(QFrame.HLine)
        hline.setStyleSheet(self._separators_stylesheet)
        self.contents_widget.setItemWidget(item, hline)

        # This is necessary to keep in sync the contents_widget and
        # pages_widget indexes.
        self.pages_widget.addWidget(QWidget(self))

        # Save separators to perform certain operations only on them
        self._separators.append(hline)

    def add_page(self, page: SidebarPage, initialize: bool = True):
        """Add page instance to the dialog."""
        if initialize:
            page.initialize()

        page.show_this_page.connect(lambda row=self.contents_widget.count():
                                    self.contents_widget.setCurrentRow(row))

        # Container widget so that we can set margins around the page
        layout = QHBoxLayout()
        layout.addWidget(page)

        # The smaller margin to the right is necessary to compensate for the
        # space added by the vertical scrollbar
        layout.setContentsMargins(27, 27, 15, 27)

        # This is necessary to show the full contents of the page in case it
        # doesn't fit vertically in the available space.
        layout.setSizeConstraint(QLayout.SetFixedSize)

        container = QWidget(self)
        container.setLayout(layout)
        container.page = page

        # Add container to a scroll area in case the page contents don't fit
        # in the dialog
        scrollarea = PageScrollArea(self)
        scrollarea.setObjectName('sidebardialog-scrollarea')
        scrollarea.setWidgetResizable(False)
        scrollarea.setWidget(container)
        scrollarea.setAlignment(Qt.AlignHCenter)
        scrollarea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.pages_widget.addWidget(scrollarea)

        # Add plugin entry item to contents widget
        item = QListWidgetItem(self.contents_widget)
        item.setText(page.get_name())
        item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)

        # In case the page doesn't have an icon
        try:
            item.setIcon(page.get_icon())
        except TypeError:
            pass

        # Set font for items
        item.setFont(self.items_font)

        # Save which pages are active in case we need to hide some
        self._active_pages[self.contents_widget.count() - 1] = True

    def number_of_pages(self):
        """Get the number of pages in the dialog."""
        return self.pages_widget.count()

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def showEvent(self, event):
        """Adjustments when the widget is shown."""
        if not self._is_shown:
            self._add_tooltips()
            self._adjust_items_margin()

            self._is_shown = True

        super().showEvent(event)

        # This is necessary to paint the separators as expected when there
        # are elided items in contents_widget.
        with signals_blocked(self):
            height = self.height()
            self.resize(self.width(), height + 1)
            self.resize(self.width(), height - 1)

    def resizeEvent(self, event):
        """
        Reimplement Qt method to perform several operations when resizing.
        """
        QDialog.resizeEvent(self, event)
        self._on_resize_event()

    # ---- Private API
    # -------------------------------------------------------------------------
    def _add_tooltips(self):
        """
        Check if it's necessary to add tooltips to the contents_widget items.
        """
        contents_width = self.contents_widget.width()
        metrics = QFontMetricsF(self.items_font)

        for i in range(self.contents_widget.count()):
            item = self.contents_widget.item(i)

            # Item width
            item_width = self.contents_widget.visualItemRect(item).width()

            # Set tooltip
            if item_width >= contents_width:
                item.setToolTip(item.text())
            else:
                # This covers the case when item_width is too close to
                # contents_width without the scrollbar being visible, which
                # can't be detected by Qt with the check above.
                scrollbar = self.contents_widget.verticalScrollBar()

                if scrollbar.isVisible():
                    if MAC:
                        # This is a crude heuristic to detect if we need to add
                        # tooltips on Mac. However, it's the best we can do
                        # (the approach for other OSes below ends up adding
                        # tooltips to all items) and it works for all our
                        # localized languages.
                        text_width = metrics.boundingRect(item.text()).width()
                        if text_width + 70 > item_width - 5:
                            item.setToolTip(item.text())
                    else:
                        if item_width > (contents_width - scrollbar.width()):
                            item.setToolTip(item.text())

    def _adjust_items_margin(self):
        """
        Adjust margins of contents_widget items depending on if its vertical
        scrollbar is visible.

        Notes
        -----
        We need to do this only in Mac because Qt doesn't account for the
        scrollbar width in most widgets.
        """
        if MAC:
            scrollbar = self.contents_widget.verticalScrollBar()
            extra_margin = (
                AppStyle.MacScrollBarWidth if scrollbar.isVisible() else 0
            )
            item_margin = (
                f'0px {self.ITEMS_MARGIN + extra_margin}px '
                f'0px {self.ITEMS_MARGIN}px'
            )

            self._contents_css['QListView::item'].setValues(
                margin=item_margin
            )

            self.contents_widget.setStyleSheet(self._contents_css.toString())

    def _adjust_separators_width(self):
        """
        Adjust the width of separators present in contents_widget depending on
        if its vertical scrollbar is visible.

        Notes
        -----
        We need to do this only in Mac because Qt doesn't set the widths
        correctly when there are elided items.
        """
        if MAC:
            scrollbar = self.contents_widget.verticalScrollBar()
            for sep in self._separators:
                if self.CONTENTS_WIDTH != 230:
                    raise ValueError(
                        "The values used here for the separators' width were "
                        "the ones reported by Qt for a contents_widget width "
                        "of 230px. Since this value changed, you need to "
                        "update them."
                    )

                # These are the values reported by Qt when CONTENTS_WIDTH = 230
                # and the interface language is English.
                if scrollbar.isVisible():
                    sep.setFixedWidth(188)
                else:
                    sep.setFixedWidth(204)

    @property
    def _main_stylesheet(self):
        """Main style for this widget."""
        # Use the preferences tabbar stylesheet as the base one and extend it.
        tabs_stylesheet = PREFERENCES_TABBAR_STYLESHEET.get_copy()
        css = tabs_stylesheet.get_stylesheet()

        # Remove border of all scroll areas for pages
        css['QScrollArea#sidebardialog-scrollarea'].setValues(
            border='0px',
        )

        # Add more spacing between QGroupBoxes than normal.
        css.QGroupBox.setValues(
            marginBottom='15px',
        )

        # Substract extra padding
        css["QToolTip"].setValues(
            paddingRight="-2px",
        )

        # Substract extra padding that comes from QLineEdit
        css["QLineEdit QToolTip"].setValues(
            padding="-2px -3px",
        )

        # This is necessary to correctly show disabled buttons in this kind of
        # dialogs (oddly QDarkstyle doesn't set this color as expected).
        css["QPushButton:disabled"].setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_4,
        )

        css["QPushButton:checked:disabled"].setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_6,
        )

        return css.toString()

    def _generate_contents_stylesheet(self):
        """Generate stylesheet for the contents widget"""
        css = qstylizer.style.StyleSheet()

        # This also sets the background color of the vertical scrollbar
        # associated to this widget
        css.setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_2
        )

        # Main style
        css.QListView.setValues(
            padding=f'{self.ITEMS_MARGIN}px 0px',
            border=f'1px solid {SpyderPalette.COLOR_BACKGROUND_2}',
        )

        # Remove border color on focus
        css['QListView:focus'].setValues(
            border=f'1px solid {SpyderPalette.COLOR_BACKGROUND_2}',
        )

        # Add margin and padding for items
        css['QListView::item'].setValues(
            padding=f'{self.ITEMS_PADDING}px',
            margin=f'0px {self.ITEMS_MARGIN}px'
        )

        # Set border radius and background color for hover, active and inactive
        # states of items
        css['QListView::item:hover'].setValues(
            borderRadius=SpyderPalette.SIZE_BORDER_RADIUS,
        )

        for state in ['item:selected:active', 'item:selected:!active']:
            css[f'QListView::{state}'].setValues(
                borderRadius=SpyderPalette.SIZE_BORDER_RADIUS,
                backgroundColor=SpyderPalette.COLOR_BACKGROUND_4
            )

        return css

    @property
    def _contents_scrollbar_stylesheet(self):
        css = qstylizer.style.StyleSheet()

        # Give border a darker color to stand out over the background
        css.setValues(
            border=f"1px solid {SpyderPalette.COLOR_BACKGROUND_5}"
        )

        return css.toString()

    @property
    def _separators_stylesheet(self):
        css = qstylizer.style.StyleSheet()

        # This makes separators stand out better over the background
        css.setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_5
        )

        return css.toString()

    @qdebounced(timeout=40)
    def _on_resize_event(self):
        """Method to run when Qt emits a resize event."""
        self._add_tooltips()
        self._adjust_items_margin()
        self._adjust_separators_width()

    def _add_pages(self):
        """Add pages to the dialog."""
        for PageClass in self.PAGE_CLASSES:
            page = PageClass(self)
            self.add_page(page)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Completion widget class."""

# Standard library imports
import html
import sys

# Third psrty imports
from qtpy.QtCore import QPoint, Qt, Signal, Slot
from qtpy.QtGui import QFontMetrics, QFocusEvent
from qtpy.QtWidgets import QListWidget, QListWidgetItem, QToolTip

# Local imports
from spyder.api.config.mixins import SpyderConfigurationAccessor
from spyder.utils.icon_manager import ima
from spyder.plugins.completion.api import CompletionItemKind
from spyder.py3compat import to_text_string
from spyder.utils.qthelpers import keyevent_to_keysequence_str
from spyder.widgets.helperwidgets import HTMLDelegate


COMPLETION_ITEM_HEIGHT = 15
COMPLETION_ITEM_WIDTH = 250
COMPLETION_DELTA_FOR_SCROLLBAR = 7 if sys.platform.startswith("linux") else 0


class CompletionWidget(QListWidget, SpyderConfigurationAccessor):
    """Completion list widget."""

    ITEM_TYPE_MAP = {
        CompletionItemKind.TEXT: 'text',
        CompletionItemKind.METHOD: 'method',
        CompletionItemKind.FUNCTION: 'function',
        CompletionItemKind.CONSTRUCTOR: 'constructor',
        CompletionItemKind.FIELD: 'field',
        CompletionItemKind.VARIABLE: 'variable',
        CompletionItemKind.CLASS: 'class',
        CompletionItemKind.INTERFACE: 'interface',
        CompletionItemKind.MODULE: 'module',
        CompletionItemKind.PROPERTY: 'property',
        CompletionItemKind.UNIT: 'unit',
        CompletionItemKind.VALUE: 'value',
        CompletionItemKind.ENUM: 'enum',
        CompletionItemKind.KEYWORD: 'keyword',
        CompletionItemKind.SNIPPET: 'snippet',
        CompletionItemKind.COLOR: 'color',
        CompletionItemKind.FILE: 'file',
        CompletionItemKind.REFERENCE: 'reference',
    }

    ICON_MAP = {}

    sig_show_completions = Signal(object)

    # Signal with the info about the current completion item documentation
    # str: completion name
    # str: completion signature/documentation,
    # QPoint: QPoint where the hint should be shown
    sig_completion_hint = Signal(str, str, QPoint)

    def __init__(self, parent, ancestor):
        super(CompletionWidget, self).__init__(ancestor)
        self.textedit = parent
        self._language = None
        self.setWindowFlags(Qt.SubWindow | Qt.FramelessWindowHint)
        self.hide()
        self.itemActivated.connect(self.item_selected)
        self.currentRowChanged.connect(self.row_changed)
        self.is_internal_console = False
        self.completion_list = None
        self.completion_position = None
        self.automatic = False
        self.current_selected_item_label = None
        self.current_selected_item_point = None
        self.display_index = []

        # Setup item rendering
        self.setItemDelegate(HTMLDelegate(self, margin=3))
        self.setMinimumWidth(COMPLETION_ITEM_WIDTH)

        # Initial item height and width
        fm = QFontMetrics(self.textedit.font())
        self.item_height = fm.height()

    def setup_appearance(self, size, font):
        """Setup size and font of the completion widget."""
        self.resize(*size)
        self.setFont(font)

    def is_empty(self):
        """Check if widget is empty."""
        if self.count() == 0:
            return True
        return False

    def show_list(self, completion_list, position, automatic):
        """Show list corresponding to position."""
        self.current_selected_item_label = None
        self.current_selected_item_point = None

        if not completion_list:
            self.hide()
            return

        self.automatic = automatic

        if position is None:
            # Somehow the position was not saved.
            # Hope that the current position is still valid
            self.completion_position = self.textedit.textCursor().position()

        elif self.textedit.textCursor().position() < position:
            # hide the text as we moved away from the position
            self.hide()
            return
        else:
            self.completion_position = position

        # Completions are handled differently for the Internal
        # console.
        if not isinstance(completion_list[0], dict):
            self.is_internal_console = True
        self.completion_list = completion_list

        # Check everything is in order
        self.update_current(new=True)

        # If update_current called close, stop loading
        if not self.completion_list:
            return

        # If only one, must be chosen if not automatic
        single_match = self.count() == 1
        if single_match and not self.automatic:
            self.item_selected(self.item(0))
            # signal used for testing
            self.sig_show_completions.emit(completion_list)
            return

        self.show()
        self.setFocus()
        self.raise_()

        if self.verticalScrollBar().isVisible():
            # This is necessary to avoid the scrollbar cropping item types
            self.setMinimumWidth(
                COMPLETION_ITEM_WIDTH - COMPLETION_DELTA_FOR_SCROLLBAR
            )
        else:
            self.setMinimumWidth(COMPLETION_ITEM_WIDTH)

        self.textedit.position_widget_at_cursor(self)

        if not self.is_internal_console:
            tooltip_point = self.rect().topRight()
            tooltip_point = self.mapToGlobal(tooltip_point)

            if self.completion_list is not None:
                for completion in self.completion_list:
                    completion['point'] = tooltip_point

        # Show hint for first completion element
        self.setCurrentRow(0)
        self.row_changed(0)

        # signal used for testing
        self.sig_show_completions.emit(completion_list)

    def set_language(self, language):
        """Set the completion language."""
        self._language = language.lower()

    def update_list(self, current_word, new=True):
        """
        Update the displayed list by filtering self.completion_list based on
        the current_word under the cursor (see check_can_complete).

        If we're not updating the list with new completions, we filter out
        textEdit completions, since it's difficult to apply them correctly
        after the user makes edits.

        If no items are left on the list the autocompletion should stop
        """
        self.clear()

        self.display_index = []
        height = self.item_height

        # This heuristics tells us if the amount of items is taller than the
        # widget's height, which makes it display the vertical scrollbar (we
        # can't check if it's visible because the widget can't be shown when
        # this function is called).
        if (
            # The +4 here is the vertical amount of padding around each item
            (height + 4) * len(self.completion_list)
            + self.horizontalScrollBar().height()
        ) > self.height() - 10:
            # This is necessary to avoid the scrollbar cropping item types
            width = COMPLETION_ITEM_WIDTH - COMPLETION_DELTA_FOR_SCROLLBAR
        else:
            # This is necessary to make the item type closer to the widget's
            # right border.
            width = COMPLETION_ITEM_WIDTH + COMPLETION_DELTA_FOR_SCROLLBAR

        for i, completion in enumerate(self.completion_list):
            if not self.is_internal_console:
                if not new and 'textEdit' in completion:
                    continue
                completion_label = completion['filterText']
            else:
                completion_label = completion[0]

            if not self.check_can_complete(completion_label, current_word):
                continue
            item = QListWidgetItem()

            if not self.is_internal_console:
                self.set_item_display(
                    item, completion, height=height, width=width)
                item.setData(Qt.UserRole, completion)
            else:
                completion_text = self.get_html_item_representation(
                    completion_label, '', height=height, width=width)
                item.setData(Qt.UserRole, completion_label)
                item.setText(completion_text)

            self.addItem(item)
            self.display_index.append(i)

        if self.count() == 0:
            self.hide()

    def _get_cached_icon(self, name):
        if name not in self.ICON_MAP:
            self.ICON_MAP[name] = ima.icon(name)
        return self.ICON_MAP[name]

    def set_item_display(self, item_widget, item_info, height, width):
        """Set item text & icons using the info available."""
        item_type = self.ITEM_TYPE_MAP.get(item_info['kind'], 'no_match')
        item_label = item_info['label']
        icon_provider = ("iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0l"
                         "EQVR42mNkYAAAAAYAAjCB0C8AAAAASUVORK5CYII=")
        img_height = height - 2
        img_width = img_height * 0.8

        icon_provider, icon_scale = item_info.get('icon', (None, 1))
        if icon_provider is not None:
            icon_height = img_height
            icon_width = icon_scale * icon_height
            icon_provider = ima.icon(icon_provider)
            icon_provider = ima.base64_from_icon_obj(
                icon_provider, icon_width, icon_height)

        item_text = self.get_html_item_representation(
            item_label, item_type, icon_provider=icon_provider,
            img_height=img_height, img_width=img_width, height=height,
            width=width)

        item_widget.setText(item_text)
        item_widget.setIcon(self._get_cached_icon(item_type))

        # Set data for accessible readers using item label and type
        # See spyder-ide/spyder#17047 and
        # https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum
        item_widget.setData(Qt.AccessibleTextRole, f"{item_label} {item_type}")

    def get_html_item_representation(self, item_completion, item_type,
                                     icon_provider=None,
                                     img_height=0,
                                     img_width=0,
                                     height=COMPLETION_ITEM_HEIGHT,
                                     width=COMPLETION_ITEM_WIDTH):
        """Get HTML representation of and item."""
        height = to_text_string(height)
        width = to_text_string(width)

        # Unfortunately, both old- and new-style Python string formatting
        # have poor performance due to being implemented as functions that
        # parse the format string.
        # f-strings in new versions of Python are fast due to Python
        # compiling them into efficient string operations, but to be
        # compatible with old versions of Python, we manually join strings.
        parts = [
            '<table width="', width, '" height="', height, '">', '<tr>',

            '<td valign="middle" style="color:' + ima.MAIN_FG_COLOR + '">',
            html.escape(item_completion).replace(' ', '&nbsp;'),
            '</td>',

            '<td valign="middle" align="right" float="right" style="color:',
            ima.MAIN_FG_COLOR, '">',
            item_type,
            '</td>',
        ]
        if icon_provider is not None:
            img_height = to_text_string(img_height)
            img_width = to_text_string(img_width)

            parts.extend([
                '<td valign="top" align="right" float="right" width="',
                img_width, '">',
                '<img src="data:image/png;base64, ', icon_provider,
                '" height="', img_height, '" width="', img_width, '"/>',
                '</td>',
            ])
        parts.extend([
            '</tr>', '</table>',
        ])

        return ''.join(parts)

    def hide(self, focus_to_parent=True):
        """Override Qt method."""
        self.completion_position = None
        self.completion_list = None
        self.clear()

        # Used to control when to give focus to its parent.
        # This is necessary to have a better fix than the initially
        # proposed for issue spyder-ide/spyder#11502.
        if focus_to_parent:
            self.textedit.setFocus()

        tooltip = getattr(self.textedit, 'tooltip_widget', None)
        if tooltip:
            # This is necessary to reuse this widget as a tooltip or hover.
            tooltip.reset_state()

            # Use this method to hide the widget immediately instead of using a
            # time, which can generate odd UX situations.
            tooltip._hide()

        QListWidget.hide(self)
        QToolTip.hideText()

    def keyPressEvent(self, event):
        """Override Qt method to process keypress."""
        text, key = event.text(), event.key()
        alt = event.modifiers() & Qt.AltModifier
        shift = event.modifiers() & Qt.ShiftModifier
        ctrl = event.modifiers() & Qt.ControlModifier
        altgr = event.modifiers() and (key == Qt.Key_AltGr)

        # Needed to properly handle Neo2 and other keyboard layouts
        # See spyder-ide/spyder#11293
        neo2_level4 = (key == 0)  # AltGr (ISO_Level5_Shift) in Neo2 on Linux
        modifier = shift or ctrl or alt or altgr or neo2_level4

        if key in (Qt.Key_Return, Qt.Key_Enter, Qt.Key_Tab):
            # Check that what was selected can be selected,
            # otherwise timing issues
            item = self.currentItem()
            if item is None:
                item = self.item(0)

            if self.is_up_to_date(item=item):
                self.item_selected(item=item)
            else:
                self.hide()
                self.textedit.keyPressEvent(event)
        elif key == Qt.Key_Escape:
            self.hide()
        elif key in (Qt.Key_Left, Qt.Key_Right) or text in ('.', ':'):
            self.hide()
            self.textedit.keyPressEvent(event)
        elif (
            key in (Qt.Key_Up, Qt.Key_Down, Qt.Key_PageUp, Qt.Key_PageDown)
            and not modifier
        ):
            self.textedit._completions_hint_idle = True
            if key == Qt.Key_Up and self.currentRow() == 0:
                self.setCurrentRow(self.count() - 1)
            elif key == Qt.Key_Down and self.currentRow() == self.count()-1:
                self.setCurrentRow(0)
            else:
                QListWidget.keyPressEvent(self, event)
        elif key in (Qt.Key_Home, Qt.Key_End):
            # This allows users to easily move to the beginning/end of the
            # current line when this widget is visible.
            # Fixes spyder-ide/spyder#19989
            self.hide()
            self.textedit.keyPressEvent(event)
        elif key == Qt.Key_Backspace:
            self.textedit.keyPressEvent(event)
            self.update_current(new=False)
        elif len(text):
            # Determine the key sequence introduced by the user to check if
            # we need to call the action associated to it in textedit. This
            # way is not necessary to hide this widget for the sequence to
            # take effect in textedit.
            # Fixes spyder-ide/spyder#19372
            if modifier:
                key_sequence = keyevent_to_keysequence_str(event)

                # Ask to save file if the user pressed the sequence for that.
                # Fixes spyder-ide/spyder#14806
                save_shortcut = self.get_conf(
                    'editor/save file', section='shortcuts')
                if key_sequence == save_shortcut:
                    self.textedit.sig_save_requested.emit()

                    # Hiding the widget reassures users that the save operation
                    # took place.
                    self.hide()

            self.textedit.keyPressEvent(event)
            self.update_current(new=False)
        elif modifier:
            self.textedit.keyPressEvent(event)
        else:
            self.hide()
            QListWidget.keyPressEvent(self, event)

    def is_up_to_date(self, item=None):
        """
        Check if the selection is up to date.
        """
        if self.is_empty():
            return False
        if not self.is_position_correct():
            return False
        if item is None:
            item = self.currentItem()
        current_word = self.textedit.get_current_word(completion=True)
        completion = item.data(Qt.UserRole)
        if isinstance(completion, dict):
            filter_text = completion['filterText']
        else:
            filter_text = completion
        return self.check_can_complete(filter_text, current_word)

    def check_can_complete(self, filter_text, current_word):
        """Check if current_word matches filter_text."""
        if not filter_text:
            return True

        if not current_word:
            return True

        return to_text_string(filter_text).lower().startswith(
                to_text_string(current_word).lower())

    def is_position_correct(self):
        """Check if the position is correct."""

        if self.completion_position is None:
            return False

        cursor_position = self.textedit.textCursor().position()

        # Can only go forward from the data we have
        if cursor_position < self.completion_position:
            return False

        completion_text = self.textedit.get_current_word_and_position(
            completion=True)

        # If no text found, we must be at self.completion_position
        if completion_text is None:
            if self.completion_position == cursor_position:
                return True
            else:
                return False

        completion_text, text_position = completion_text
        completion_text = to_text_string(completion_text)

        # The position of text must compatible with completion_position
        if not text_position <= self.completion_position <= (
                text_position + len(completion_text)):
            return False

        return True

    def update_current(self, new=False):
        """
        Update the displayed list.
        """
        if not self.is_position_correct():
            self.hide()
            return

        current_word = self.textedit.get_current_word(completion=True)
        self.update_list(current_word, new=new)

    def focusOutEvent(self, event):
        """Override Qt method."""

        # Type check: Prevent error in PySide where 'event' may be of type
        # QtGui.QPainter (for whatever reason).
        if type(event) is not QFocusEvent:
            return

        event.ignore()
        # Don't hide it on Mac when main window loses focus because
        # keyboard input is lost.
        # Fixes spyder-ide/spyder#1318.
        if sys.platform == "darwin":
            if event.reason() != Qt.ActiveWindowFocusReason:
                self.hide()
        else:
            # Avoid an error when running tests that show
            # the completion widget
            try:
                self.hide()
            except RuntimeError:
                pass

    def item_selected(self, item=None):
        """Perform the item selected action."""
        if item is None:
            item = self.currentItem()

        if item is not None and self.completion_position is not None:
            self.textedit.insert_completion(item.data(Qt.UserRole),
                                            self.completion_position)
        self.hide()

    def _get_insert_text(self, item):
        if 'textEdit' in item:
            insert_text = item['textEdit']['newText']
        else:
            insert_text = item['insertText']

            # Split by starting $ or language specific chars
            chars = ['$']
            if self._language == 'python':
                chars.append('(')

            for ch in chars:
                insert_text = insert_text.split(ch)[0]
        return insert_text

    def request_completion_hint(self, row=None):
        """Request LSP for completion hint"""
        self.current_selected_item_label = None
        self.current_selected_item_point = None

        if not self.completion_list:
            return
        if row is None:
            row = self.currentRow()
        if row < 0 or len(self.completion_list) <= row:
            return

        item = self.completion_list[row]
        if 'point' not in item:
            return

        self.current_selected_item_label = item['label']
        self.current_selected_item_point = item['point']

        # Ask LSP for completion hint
        if hasattr(self.textedit, 'resolve_completion_item'):
            if item.get('resolve', False):
                to_resolve = item.copy()
                to_resolve.pop('point')
                to_resolve.pop('resolve')
                self.textedit.resolve_completion_item(to_resolve)
            else:
                self.textedit.hide_tooltip()

    def augment_completion_info(self, item):
        """Augment completion info with hints that come from the server."""
        if self.current_selected_item_label == item['label']:
            insert_text = self._get_insert_text(item)

            if isinstance(item['documentation'], dict):
                item['documentation'] = item['documentation']['value']

            self.sig_completion_hint.emit(
                insert_text,
                item['documentation'],
                self.current_selected_item_point)

    @Slot(int)
    def row_changed(self, row):
        """Actions to take when the row has changed."""
        self.request_completion_hint(row)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Profiler run executor configurations."""

# Standard library imports
import os.path as osp

# Third-party imports
from qtpy.compat import getexistingdirectory
from qtpy.QtWidgets import (
    QGroupBox, QVBoxLayout, QGridLayout, QCheckBox, QLineEdit)

# Local imports
from spyder.api.translations import _
from spyder.plugins.run.api import (
    RunExecutorConfigurationGroup, Context, RunConfigurationMetadata)
from spyder.utils.misc import getcwd_or_home


class ProfilerPyConfigurationGroup(RunExecutorConfigurationGroup):
    """External console Python run configuration options."""

    def __init__(self, parent, context: Context, input_extension: str,
                 input_metadata: RunConfigurationMetadata):
        super().__init__(parent, context, input_extension, input_metadata)

        self.dir = None

        # --- General settings ----
        common_group = QGroupBox(_("File settings"))
        common_layout = QGridLayout(common_group)

        self.clo_cb = QCheckBox(_("Command line options:"))
        common_layout.addWidget(self.clo_cb, 0, 0)
        self.clo_edit = QLineEdit(self)
        self.clo_edit.setMinimumWidth(300)
        self.clo_cb.toggled.connect(self.clo_edit.setEnabled)
        self.clo_edit.setEnabled(False)
        common_layout.addWidget(self.clo_edit, 0, 1)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(common_group)
        layout.addStretch(100)

    def select_directory(self):
        """Select directory"""
        basedir = str(self.wd_edit.text())
        if not osp.isdir(basedir):
            basedir = getcwd_or_home()
        directory = getexistingdirectory(self, _("Select directory"), basedir)
        if directory:
            self.wd_edit.setText(directory)
            self.dir = directory

    @staticmethod
    def get_default_configuration() -> dict:
        return {
            'args_enabled': False,
            'args': ''
        }

    def set_configuration(self, config: dict):
        args_enabled = config['args_enabled']
        args = config['args']

        self.clo_cb.setChecked(args_enabled)
        self.clo_edit.setText(args)

    def get_configuration(self) -> dict:
        return {
            'args_enabled': self.clo_cb.isChecked(),
            'args': self.clo_edit.text(),
        }



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for qcookiecutter widget.
"""

# Standard library imports
import os
from unittest.mock import Mock

# Third party imports
import pytest

# Local imports
from spyder.plugins.projects.widgets.qcookiecutter import CookiecutterWidget


@pytest.fixture
def coookie_widget(qtbot):
    """Set up CookieCutter Widget."""
    widget = CookiecutterWidget(None)
    qtbot.addWidget(widget)
    return widget


def test_cookiecutter_widget_empty(coookie_widget):
    assert len(coookie_widget._widgets) == 0
    assert len(coookie_widget.get_values()) == 3

    coookie_widget.setup({})
    assert len(coookie_widget._widgets) == 0
    assert len(coookie_widget.get_values()) == 3


@pytest.mark.parametrize("option,value", [
    ("opt", "y"),
    ("opt", "yes"),
    ("opt", "true"),
    ("opt", "YES"),
    ("opt", "True"),
])
def test_cookiecutter_widget_checkbox_yes(coookie_widget, option, value):
    coookie_widget.setup({option: value})
    label, widget = coookie_widget._widgets[option]
    assert len(coookie_widget._widgets) == 1
    assert label == option.capitalize()
    assert widget.isChecked()
    assert widget.get_value() == value


@pytest.mark.parametrize("option,value", [
    ("opt", "n"),
    ("opt", "no"),
    ("opt", "false"),
    ("opt", "NO"),
    ("opt", "False"),
])
def test_cookiecutter_widget_checkbox_no(coookie_widget, option, value):
    coookie_widget.setup({option: value})
    label, widget = coookie_widget._widgets[option]
    assert len(coookie_widget._widgets) == 1
    assert label == option.capitalize()
    assert not widget.isChecked()
    assert widget.get_value() == value


@pytest.mark.parametrize("option,value", [
    ("opt", ["1", "2", "3"]),
])
def test_cookiecutter_widget_list(coookie_widget, option, value):
    coookie_widget.setup({option: value})
    label, widget = coookie_widget._widgets[option]
    assert len(coookie_widget._widgets) == 1
    assert label == option.capitalize()
    assert widget.get_value() == value[0]


@pytest.mark.parametrize("option,value", [
    ("opt", {"1": [1, 2], "2": [3, 4]}),
])
def test_cookiecutter_widget_dict(coookie_widget, option, value):
    coookie_widget.setup({option: value})
    label, widget = coookie_widget._widgets[option]
    assert len(coookie_widget._widgets) == 1
    assert label == option.capitalize()
    assert widget.get_value() == {"1": value["1"]}


@pytest.mark.parametrize("option,value", [
    ("_nope", "nothing"),
    ("__nope_2", "nothing"),
])
def test_cookiecutter_widget_private_variables(coookie_widget, option, value):
    coookie_widget.setup({option: value})
    assert len(coookie_widget._widgets) == 0
    assert len(coookie_widget.get_values()) == 4


def test_cookiecutter_widget_render(coookie_widget):
    coookie_widget.setup({
        "opt_1": "test",
        "opt_2": "{{ cookiecutter.opt_1 }}",
    })
    ows = coookie_widget._widgets
    assert ows["opt_2"][1].get_value() == ows["opt_1"][1].get_value()


def test_cookiecutter_widget_no_render(coookie_widget):
    coookie_widget.setup({
        "opt_1": "test",
        "opt_2": "{{ cookiecutter.opt_1 }}",
        "_opt_3": "{{ cookiecutter.opt_1 }}",
        "__opt_4": "{{ cookiecutter.opt_1 }}",
    })
    ows = coookie_widget.get_values()
    assert ows["_opt_3"] == ows["_opt_3"]
    assert ows["__opt_4"] == ows["__opt_4"]


def test_cookiecutter_widget_validate_passes(qtbot, coookie_widget):
    coookie_widget.setup({
        "opt_1": "test",
    })
    coookie_widget.set_pre_gen_code('''
import sys
sys.exit(0)
''')
    with qtbot.waitSignal(coookie_widget.sig_validated) as blocker:
        coookie_widget.validate()

    assert blocker.args == [0, ""]


def test_cookiecutter_widget_validate_fails(qtbot, coookie_widget):
    coookie_widget.setup({
        "opt_1": "test",
    })
    coookie_widget.set_pre_gen_code('''
import sys
print('ERROR!')  # spyder: test-skip
sys.exit(1)
''')
    with qtbot.waitSignal(coookie_widget.sig_validated) as blocker:
        coookie_widget.validate()

    assert blocker.args == [1, "ERROR! "]


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for findinfiles.py
"""

# Test library imports
import os
import os.path as osp
from unittest.mock import MagicMock

# Third party imports
from flaky import flaky
import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QMessageBox

# Local imports
from spyder.config.base import running_in_ci
from spyder.config.manager import CONF
from spyder.plugins.findinfiles.widgets.main_widget import FindInFilesWidget
from spyder.plugins.findinfiles.widgets.combobox import (
    SearchInComboBox,
    SearchInComboBoxItems
)
from spyder.plugins.findinfiles.widgets.search_thread import SearchThread
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import APP_STYLESHEET


LOCATION = osp.realpath(osp.join(os.getcwd(), osp.dirname(__file__)))
NONASCII_DIR = osp.join(LOCATION, "èáïü Øαôå 字分误")
if not osp.exists(NONASCII_DIR):
    os.makedirs(NONASCII_DIR)


def process_search_results(results):
    """
    Transform result representation from the output of the widget to the
    test framework comparison representation.
    """
    matches = {}
    for result in results.values():
        file, line, col, __ = result
        filename = osp.basename(file)
        if filename not in matches:
            matches[filename] = []
        matches[filename].append((line, col))
        matches[filename] = sorted(matches[filename])
    return matches


@pytest.fixture
def findinfiles(qtbot, request):
    """Set up find in files widget."""
    if getattr(request, 'param', False):
        param = request.param
    else:
        param = None

    plugin_mock = MagicMock()
    plugin_mock.CONF_SECTION = 'find_in_files'
    if param:
        prev_values = {}
        for param_name in param:
            value = param[param_name]
            prev_values[param_name] = CONF.get('find_in_files', param_name)
            CONF.set('find_in_files', param_name, value)

        widget = FindInFilesWidget('find_in_files', plugin=plugin_mock)
        widget._setup()
        widget.setup()

        def teardown():
            for param_name in prev_values:
                value = prev_values[param_name]
                CONF.set('find_in_files', param_name, value)

        request.addfinalizer(teardown)
    else:
        widget = FindInFilesWidget('find_in_files', plugin=plugin_mock)
        widget._setup()
        widget.setup()

    widget.resize(640, 480)
    widget.setStyleSheet(str(APP_STYLESHEET))
    qtbot.addWidget(widget)
    widget.show()
    return widget


@pytest.fixture
def searchin_combobox(qtbot, request):
    """Set up SearchInComboBox combobox."""
    from spyder.plugins.findinfiles.widgets import combobox

    if getattr(request, 'param', False):
        param = request.param
    else:
        param = None

    if param and param.get('max_history_path'):
        combobox.MAX_PATH_HISTORY = param.get('max_history_path')

    external_path_history = [
            LOCATION,
            osp.dirname(LOCATION),
            osp.dirname(osp.dirname(LOCATION)),
            osp.dirname(osp.dirname(osp.dirname(LOCATION))),
            osp.dirname(LOCATION),
            osp.join(LOCATION, 'path_that_does_not_exist')
    ]
    searchin_combobox = SearchInComboBox(external_path_history)
    qtbot.addWidget(searchin_combobox)
    return searchin_combobox


def expected_results():
    results = {'spam.txt': [(1, 0), (1, 5), (3, 22)],
               'spam.py': [(2, 7), (5, 1), (7, 12)],
               'spam.cpp': [(2, 9), (6, 15), (8, 2), (11, 4),
                            (11, 10), (13, 12)]
               }
    return results


def expected_case_unsensitive_results():
    results = {'spam.txt': [(1, 10)],
               'ham.txt': [(1, 0), (1, 10), (3, 0), (4, 0),
                           (5, 4), (9, 0), (10, 0)]}
    return results


@flaky(max_runs=5)
def test_find_in_files_search(findinfiles, qtbot):
    """
    Test the find in files utility by searching a string located on a set of
    known files.

    The results of the test should be equal to the expected search result
    values.
    """
    findinfiles.set_search_text("spam")
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.find()
    blocker = qtbot.waitSignal(findinfiles.sig_finished)
    blocker.wait()
    matches = process_search_results(findinfiles.result_browser.data)
    assert expected_results() == matches


@pytest.mark.parametrize('findinfiles',
                         [{'exclude': r"\.py$", 'exclude_regexp': True}],
                         indirect=True)
def test_exclude_extension_regex(findinfiles, qtbot):
    findinfiles.set_search_text("spam")
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.find()
    blocker = qtbot.waitSignal(findinfiles.sig_finished)
    blocker.wait()
    matches = process_search_results(findinfiles.result_browser.data)
    files_filtered = True
    for file in matches:
        filename, ext = osp.splitext(file)
        if ext == '.py':
            files_filtered = False
            break
    assert files_filtered


@pytest.mark.parametrize('findinfiles',
                         [{'exclude': "*.py", 'exclude_regexp': False}],
                         indirect=True)
def test_exclude_extension_string(findinfiles, qtbot):
    findinfiles.set_search_text("spam")
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.find()
    blocker = qtbot.waitSignal(findinfiles.sig_finished)
    blocker.wait()
    matches = process_search_results(findinfiles.result_browser.data)
    files_filtered = True
    for file in matches:
        filename, ext = osp.splitext(file)
        if ext == '.py':
            files_filtered = False
            break
    assert files_filtered


@pytest.mark.parametrize('findinfiles',
                         [{'exclude': "", 'exclude_regexp': True}],
                         indirect=True)
def test_exclude_extension_empty_regex(findinfiles, qtbot):
    findinfiles.set_search_text("spam")
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.find()
    blocker = qtbot.waitSignal(findinfiles.sig_finished)
    blocker.wait()
    matches = process_search_results(findinfiles.result_browser.data)
    assert expected_results() == matches


@pytest.mark.parametrize('findinfiles',
                         [{'exclude': "", 'exclude_regexp': False}],
                         indirect=True)
def test_exclude_extension_string_no_regexp(findinfiles, qtbot):
    findinfiles.set_search_text("spam")
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.find()
    blocker = qtbot.waitSignal(findinfiles.sig_finished)
    blocker.wait()
    matches = process_search_results(findinfiles.result_browser.data)
    assert expected_results() == matches


@pytest.mark.parametrize('findinfiles',
                         [{'exclude': "*.py, *.cpp", 'exclude_regexp': False}],
                         indirect=True)
def test_exclude_extension_multiple_string(findinfiles, qtbot):
    findinfiles.set_search_text("spam")
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.find()
    blocker = qtbot.waitSignal(findinfiles.sig_finished)
    blocker.wait()
    matches = process_search_results(findinfiles.result_browser.data)
    files_filtered = True
    for file in matches:
        filename, ext = osp.splitext(file)
        if ext in ['.py', '.cpp']:
            files_filtered = False
            break
    assert files_filtered


@pytest.mark.parametrize("line_input", ['nnnnn', 'ñandú'])
def test_truncate_result_with_different_input(findinfiles, qtbot, line_input):
    """
    Issue: 6218 - checking if truncate_result raise UnicodeDecodeError
    """

    # with
    slice_start = 1
    slice_end = 2

    line_input_expected = line_input

    expected_result = (
        f'<span style="color:{SpyderPalette.COLOR_TEXT_1}">'
        f'{line_input_expected[:slice_start]}'
        f'<span style="background-color:{SpyderPalette.COLOR_OCCURRENCE_4}">'
        f'{line_input_expected[slice_start:slice_end]}</span>'
        f'{line_input_expected[slice_end:]}</span>'
    )

    # when
    thread = SearchThread(None, '', text_color=SpyderPalette.COLOR_TEXT_1)
    truncated_line = thread.truncate_result(line_input, slice_start,
                                            slice_end)
    # then
    assert truncated_line['formatted_text'] == expected_result


@pytest.mark.parametrize('findinfiles',
                         [{'case_sensitive': False}],
                         indirect=True)
def test_case_unsensitive_search(findinfiles, qtbot):
    findinfiles.set_search_text('ham')
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.find()
    blocker = qtbot.waitSignal(findinfiles.sig_finished)
    blocker.wait()
    matches = process_search_results(findinfiles.result_browser.data)
    print(matches)
    assert expected_case_unsensitive_results() == matches


@pytest.mark.parametrize('findinfiles',
                         [{'case_sensitive': True}],
                         indirect=True)
def test_case_sensitive_search(findinfiles, qtbot):
    findinfiles.set_search_text('HaM')
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.find()
    blocker = qtbot.waitSignal(findinfiles.sig_finished)
    blocker.wait()
    matches = process_search_results(findinfiles.result_browser.data)
    print(matches)
    assert matches == {'ham.txt': [(9, 0)]}


@pytest.mark.parametrize('findinfiles',
                         [{'search_text_regexp': True}],
                         indirect=True)
def test_search_regexp_error(findinfiles, qtbot):
    findinfiles.set_search_text("\\")
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.find()
    tooltip = findinfiles.search_text_edit.toolTip()
    assert findinfiles.REGEX_ERROR in tooltip


@pytest.mark.parametrize('findinfiles',
                         [{'exclude': "\\", 'exclude_regexp': True}],
                         indirect=True)
def test_exclude_regexp_error(findinfiles, qtbot):
    findinfiles.set_search_text("foo")
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.find()
    tooltip = findinfiles.exclude_pattern_edit.toolTip()
    assert findinfiles.REGEX_ERROR in tooltip


@flaky(max_runs=5)
@pytest.mark.skipif(not running_in_ci(), reason="Only works on CIs")
def test_no_empty_file_items(findinfiles, qtbot):
    """
    Test that a search that hits the max number of results doesn't generate
    empty file items.

    This is a regression test for issue spyder-ide/spyder#16256
    """
    max_results = 6
    findinfiles.set_search_text("spam")
    findinfiles.set_directory(osp.join(LOCATION, "data"))
    findinfiles.set_conf('max_results', max_results)

    with qtbot.waitSignal(findinfiles.sig_max_results_reached):
        findinfiles.find()

    # Assert that the results all come from the expected files and that there
    # are the correct number of them.  (We do not list the exact results
    # expected because os.walk (used by findinfiles) gives an arbitrary file
    # ordering.)
    spamfiles = set(['spam.py', 'spam.txt', 'spam.cpp'])
    find_results = process_search_results(findinfiles.result_browser.data)
    assert set(find_results.keys()).issubset(spamfiles)
    assert sum(len(finds) for finds in find_results.values()) == max_results

    # Assert that the files with results are exactly the same as those
    # displayed in the results browser.
    files_with_results = set(
        [v[0] for v in findinfiles.result_browser.data.values()]
    )
    displayed_files = set(findinfiles.result_browser.files.keys())
    assert files_with_results == displayed_files


# ---- Tests for SearchInComboBox

def test_add_external_paths(searchin_combobox, mocker):
    """
    Test that the external_path_history is added correctly to the
    combobox and test that adding new external path to the combobox
    with the QFileDialog is working as expected.
    """
    searchin_combobox.show()

    # Assert that the external_path_history was added correctly to the
    # combobox
    expected_results = [
            LOCATION,
            osp.dirname(osp.dirname(LOCATION)),
            osp.dirname(osp.dirname(osp.dirname(LOCATION))),
            osp.dirname(LOCATION)
            ]
    assert (
        searchin_combobox.count()
        == len(expected_results) + SearchInComboBoxItems.ExternalPaths
    )
    assert searchin_combobox.get_external_paths() == expected_results
    for i, expected_result in enumerate(expected_results):
        assert expected_result == searchin_combobox.itemText(
            i + SearchInComboBoxItems.ExternalPaths
        )

    # Add a new external path to the combobox. The new path is added at the
    # end of the combobox.
    new_path = NONASCII_DIR
    mocker.patch(
        'spyder.plugins.findinfiles.widgets.combobox.getexistingdirectory',
        return_value=new_path
    )
    searchin_combobox.setCurrentIndex(
        SearchInComboBoxItems.SelectAnotherDirectory
    )

    expected_results.append(new_path)
    assert (
        searchin_combobox.count()
        == len(expected_results) + SearchInComboBoxItems.ExternalPaths
    )
    assert searchin_combobox.get_external_paths() == expected_results
    assert searchin_combobox.currentIndex() == searchin_combobox.count()-1

    # Add an external path that is already listed in the combobox. In this
    # case, the new path is removed from the list and is added back at the end.
    new_path = LOCATION
    mocker.patch(
        'spyder.plugins.findinfiles.widgets.combobox.getexistingdirectory',
        return_value=new_path
    )
    searchin_combobox.setCurrentIndex(
        SearchInComboBoxItems.SelectAnotherDirectory
    )

    expected_results.pop(0)
    expected_results.append(new_path)
    assert (
        searchin_combobox.count()
        == len(expected_results) + SearchInComboBoxItems.ExternalPaths
    )
    assert searchin_combobox.get_external_paths() == expected_results
    assert searchin_combobox.currentIndex() == searchin_combobox.count()-1

    # Cancel the action of adding a new external path. In this case, the
    # expected results do not change.
    mocker.patch(
        'spyder.plugins.findinfiles.widgets.combobox.getexistingdirectory',
        return_value=''
    )
    searchin_combobox.setCurrentIndex(
        SearchInComboBoxItems.SelectAnotherDirectory
    )

    assert (
        searchin_combobox.count()
        == len(expected_results) + SearchInComboBoxItems.ExternalPaths
    )
    assert searchin_combobox.get_external_paths() == expected_results
    assert searchin_combobox.currentIndex() == SearchInComboBoxItems.Cwd


def test_clear_this_list(searchin_combobox, mocker):
    """
    Test the option in the searchin combobox to clear the list of
    external paths.
    """
    searchin_combobox.show()

    # Cancel the Clear the list action and assert the result.
    mocker.patch.object(QMessageBox, 'question', return_value=QMessageBox.No)
    searchin_combobox.setCurrentIndex(SearchInComboBoxItems.ClearList)

    expected_results = [
            LOCATION,
            osp.dirname(osp.dirname(LOCATION)),
            osp.dirname(osp.dirname(osp.dirname(LOCATION))),
            osp.dirname(LOCATION)
            ]
    assert (
        searchin_combobox.count()
        == len(expected_results) + SearchInComboBoxItems.ExternalPaths
    )
    assert searchin_combobox.get_external_paths() == expected_results
    assert searchin_combobox.currentIndex() == SearchInComboBoxItems.Cwd

    # Clear the list of external paths and assert that the list of
    # external paths is empty.
    mocker.patch.object(QMessageBox, 'question', return_value=QMessageBox.Yes)
    searchin_combobox.setCurrentIndex(SearchInComboBoxItems.ClearList)

    assert searchin_combobox.count() == SearchInComboBoxItems.ExternalPaths
    assert searchin_combobox.get_external_paths() == []
    assert searchin_combobox.currentIndex() == SearchInComboBoxItems.Cwd


def test_delete_path(searchin_combobox, qtbot, mocker):
    """
    Test that the selected external path in the combobox view is removed
    correctly when the Delete key is pressed.
    """
    searchin_combobox.show()

    expected_results = [
        LOCATION,
        osp.dirname(osp.dirname(LOCATION)),
        osp.dirname(osp.dirname(osp.dirname(LOCATION))),
        osp.dirname(LOCATION)
    ]

    searchin_combobox.showPopup()
    assert searchin_combobox.currentIndex() == SearchInComboBoxItems.Cwd
    assert (
        searchin_combobox.view().currentIndex().row()
        == SearchInComboBoxItems.Cwd
    )

    # Assert that the delete action does nothing when the selected item in
    # the combobox view is not an external path.
    for i in range(SearchInComboBoxItems.ExternalPaths):
        searchin_combobox.view().setCurrentIndex(
            searchin_combobox.model().index(i, 0))
        qtbot.keyPress(searchin_combobox.view(), Qt.Key_Delete)

    assert (
        searchin_combobox.count()
        == len(expected_results) + SearchInComboBoxItems.ExternalPaths
    )
    assert searchin_combobox.get_external_paths() == expected_results
    assert searchin_combobox.currentIndex() == SearchInComboBoxItems.Cwd

    # Delete the first external path in the list.
    searchin_combobox.view().setCurrentIndex(
        searchin_combobox.model().index(SearchInComboBoxItems.ExternalPaths, 0)
    )
    qtbot.keyPress(searchin_combobox.view(), Qt.Key_Delete)

    expected_results.pop(0)
    assert (
        searchin_combobox.count()
        == len(expected_results) + SearchInComboBoxItems.ExternalPaths
    )
    assert searchin_combobox.get_external_paths() == expected_results
    assert (
        searchin_combobox.currentIndex() == SearchInComboBoxItems.ExternalPaths
    )
    assert (
        searchin_combobox.view().currentIndex().row()
        == SearchInComboBoxItems.ExternalPaths
    )

    # Delete the second external path in the remaining list.
    searchin_combobox.view().setCurrentIndex(
        searchin_combobox.model().index(
            SearchInComboBoxItems.ExternalPaths + 1, 0
        )
    )
    qtbot.keyPress(searchin_combobox.view(), Qt.Key_Delete)

    expected_results.pop(1)
    assert (
        searchin_combobox.count()
        == len(expected_results) + SearchInComboBoxItems.ExternalPaths
    )
    assert searchin_combobox.get_external_paths() == expected_results
    assert (
        searchin_combobox.currentIndex()
        == SearchInComboBoxItems.ExternalPaths + 1
    )
    assert (
        searchin_combobox.view().currentIndex().row()
        == SearchInComboBoxItems.ExternalPaths + 1
    )

    # Delete the last external path in the list.
    searchin_combobox.view().setCurrentIndex(
            searchin_combobox.model().index(searchin_combobox.count()-1, 0))
    qtbot.keyPress(searchin_combobox.view(), Qt.Key_Delete)

    expected_results.pop()
    assert (
        searchin_combobox.count()
        == len(expected_results) + SearchInComboBoxItems.ExternalPaths
    )
    assert searchin_combobox.get_external_paths() == expected_results
    assert searchin_combobox.currentIndex() == searchin_combobox.count()-1
    assert (searchin_combobox.view().currentIndex().row() ==
            searchin_combobox.count()-1)

    # Delete the last remaining external path in the list.
    searchin_combobox.view().setCurrentIndex(
        searchin_combobox.model().index(SearchInComboBoxItems.ExternalPaths, 0)
    )
    qtbot.keyPress(searchin_combobox.view(), Qt.Key_Delete)

    assert searchin_combobox.count() == SearchInComboBoxItems.ExternalPaths
    assert searchin_combobox.get_external_paths() == []
    assert searchin_combobox.currentIndex() == SearchInComboBoxItems.Cwd
    assert (
        searchin_combobox.view().currentIndex().row()
        == SearchInComboBoxItems.Cwd
    )


def test_set_project_path(findinfiles, qtbot):
    """
    Test setting the project path of the SearchInComboBox from the
    FindInFilesWidget.
    """
    path_selection_combo = findinfiles.path_selection_combo
    findinfiles.show()

    assert (
        path_selection_combo.model().item(
            SearchInComboBoxItems.Project, 0
        ).isEnabled()
        is False
    )
    assert findinfiles.project_path == ""
    assert path_selection_combo.project_path == ""

    # Set the project path to an existing directory. For the purpose of this
    # test, it doesn't need to be a valid Spyder project path.
    project_path = NONASCII_DIR
    findinfiles.set_project_path(project_path)
    assert (
        path_selection_combo.model().item(
            SearchInComboBoxItems.Project, 0
        ).isEnabled()
        is True
    )
    assert findinfiles.project_path == project_path
    assert path_selection_combo.project_path == project_path

    # Disable the project path search in the widget.
    path_selection_combo.setCurrentIndex(SearchInComboBoxItems.Project)
    findinfiles.disable_project_search()
    assert (
        path_selection_combo.model().item(
            SearchInComboBoxItems.Project, 0
        ).isEnabled()
        is False
    )
    assert findinfiles.project_path == ""
    assert path_selection_combo.project_path == ""
    assert path_selection_combo.currentIndex() == SearchInComboBoxItems.Cwd


@pytest.mark.parametrize('findinfiles',
                         [{'path_history': [
                             LOCATION,
                             osp.dirname(LOCATION),
                             osp.dirname(osp.dirname(LOCATION)),
                             NONASCII_DIR]}],
                         indirect=True)
def test_current_search_path(findinfiles, qtbot):
    """
    Test that the expected search path is returned for the corresponding
    option selected in the SearchInComboBox. This test is done using the
    FindInFilesWidget.
    """
    external_paths = [
        LOCATION,
        osp.dirname(LOCATION),
        osp.dirname(osp.dirname(LOCATION)),
        NONASCII_DIR
    ]

    path_selection_combo = findinfiles.path_selection_combo
    findinfiles.show()

    # Set the project, file, and spyder path of the SearchInComboBox.
    # For the purpose of this test, the project path doesn't need to be a
    # valid Spyder project path.

    directory = NONASCII_DIR
    project_path = NONASCII_DIR
    file_path = osp.join(directory, "spam.py")

    findinfiles.set_directory(directory)
    assert findinfiles.path == directory
    assert path_selection_combo.path == directory

    findinfiles.set_project_path(project_path)
    assert findinfiles.project_path == project_path
    assert path_selection_combo.project_path == project_path

    findinfiles.set_file_path(file_path)
    assert findinfiles.file_path == file_path
    assert path_selection_combo.file_path == file_path

    # Assert that the good path is returned for each option selected.

    # Test for the current working directory :
    path_selection_combo.setCurrentIndex(SearchInComboBoxItems.Cwd)
    assert path_selection_combo.get_current_searchpath() == directory
    assert path_selection_combo.is_file_search() is False
    # Test for the project path :
    path_selection_combo.setCurrentIndex(SearchInComboBoxItems.Project)
    assert path_selection_combo.get_current_searchpath() == project_path
    assert path_selection_combo.is_file_search() is False
    # Test for the file path :
    path_selection_combo.setCurrentIndex(SearchInComboBoxItems.File)
    assert path_selection_combo.get_current_searchpath() == file_path
    assert path_selection_combo.is_file_search() is True
    # Test for the external file path :
    for i, path in enumerate(external_paths):
        path_selection_combo.setCurrentIndex(
            SearchInComboBoxItems.ExternalPaths + i
        )
        assert path_selection_combo.get_current_searchpath() == path
        assert path_selection_combo.is_file_search() is False


@pytest.mark.parametrize('searchin_combobox',
                         [{'max_history_path': 3}],
                         indirect=True)
def test_max_history(searchin_combobox):
    """
    Test that the specified maximum number of external path is observed.
    """
    searchin_combobox.show()

    # In this case, the first path of the external_path_history was removed to
    # respect the MAX_PATH_HISTORY of 3.
    expected_results = [
            osp.dirname(osp.dirname(LOCATION)),
            osp.dirname(osp.dirname(osp.dirname(LOCATION))),
            osp.dirname(LOCATION)
    ]
    assert (
        searchin_combobox.count()
        == len(expected_results) + SearchInComboBoxItems.ExternalPaths
    )
    assert searchin_combobox.get_external_paths() == expected_results


def test_max_results(findinfiles, qtbot):
    """Test max results correspond to expected results."""
    value = 2
    findinfiles.set_max_results(value)
    findinfiles.set_search_text("spam")
    findinfiles.set_directory(osp.join(LOCATION, "data"))

    findinfiles.find()
    blocker = qtbot.waitSignal(findinfiles.sig_max_results_reached)
    blocker.wait()

    print(len(findinfiles.result_browser.data), value)
    assert len(findinfiles.result_browser.data) == value

    # Restore defaults
    findinfiles.set_max_results(1000)


@flaky(max_runs=5)
def test_find_in_single_file(findinfiles, qtbot):
    """
    Test that find in files works for a single file.

    This a regression test for issues spyder-ide/spyder#17443 and
    spyder-ide/spyder#20964.
    """
    findinfiles.set_search_text("spam")
    findinfiles.set_file_path(osp.join(LOCATION, "data", 'spam.txt'))
    findinfiles.path_selection_combo.setCurrentIndex(SearchInComboBoxItems.File)

    with qtbot.waitSignal(findinfiles.sig_finished):
        findinfiles.find()

    matches = process_search_results(findinfiles.result_browser.data)
    assert list(matches.keys()) == ['spam.txt']
    assert expected_results()['spam.txt'] == matches['spam.txt']


if __name__ == "__main__":
    pytest.main(['-x', osp.basename(__file__), '-v', '-rw'])
